from sparknlp_jsl.annotator import ContextualParserModel,ContextualParserApproach

class ContextualParser:
    @staticmethod
    def get_default_model():
        return ContextualParserModel() \
            .setInputCols(["document","token"]) \
            .setOutputCol("parsed_context")

    @staticmethod
    def get_trainable_model():
        return ContextualParserApproach() \
            .setInputCols(["document","token"]) \
            .setOutputCol("parsed_context")




import nlu

b = nlu.load('bert')
data  = 'I love KERAS + NLU <3'
df = b.predict(data)
# CKL mat means th
np_bert_CKL_mat = df['word_embeddings@small_bert_L2_128_embeddings'].values
"""
 Tensor Docs https://www.tensorflow.org/guide/tensor 
 Ragged Tensor https://www.tensorflow.org/guide/tensor#ragged_tensors 
 Sparse Tensor https://www.tensorflow.org/guide/tensor#sparse_tensors

Righ now NLU returns NP arrays. One sucy thing is, if you take a column of Word Vectors in 


"""

from keras import backend as K
from keras.layers import Layer
import tensorflow as tf
import tensorflow as tf
import pandas as pd

def deduct_vector_dim(pipe, vec_col='word_embeddings@small_bert_L2_128_embeddings'):


class WordEmbeddingsLayer(tf.keras.layers.Layer):
    """ Tokenized and embed UNTOKENIZED RAW TEXT
    This layer has 2 modes
    1. If a Tensor with 1 String is input, the String will be tokenized and a tensor one vector per word is returned
    2. If a Tensor with more than 1 String is input (dim=S>1), then the layer will view this as a vector of independent Strings.
       Each String will the tokenized and embedded and a Matrix of  shape SxD will be returned
       where S is num Strings and D is dim of embedding"""
    def __init__(self, n_units=8):
        super().__init__()
        self.n_units = n_units
    def build(self, input_shape):
        self.nlu_model = nlu.load('bert')
        self.embed_key = 'word_embeddings@small_bert_L2_128_embeddings'

    def call(self, inputs, training=False):
        """
        Input Options, W=Words, N=Num Rows aka Documents
        AUTO-DETECT-TOKEN-MODE --- WORD EMBEDDINGS
        1. Input 1 Document per Tensor, 1xW --> Tensor Shape[0] = 1
        2. Input N Document per Tensor, NxW --> Tensor Shape[0] >1
        """
        pd_series       = pd.Series(tensor_of_strings.numpy())
        pd_embed_series = self.nlu_model.predict(pd_series, output_level='document')[self.embed_key].values

        # return tf.ones(1)
        return tf.matmul(inputs,inputs)


tensor_of_strings = tf.constant(["Gray wolf","Quick brown fox","Lazy dog"])
ragged_tensor  = tf.strings.split(tensor_of_strings)
layer = DoubleLinearLayer()
# x = tf.ones((3, 100))
# res = layer(x)

# TF->NP->PDS->SDF
pd_series = pd.Series(tensor_of_strings.numpy())
res = layer(tensor_of_strings)



print('yp',res)
print('yp',res)
print('yp',res)
print('yp',res)
data = 'yo'


"""Collection of methods to substitute cols of open source component results
For most annotators we can re-use a common name.
Some make an exception and need special names derives, i.e. ClassifierDl/MultiClassifierDl/ etc.. which should be derived from nlu_ref.

"""
import logging
logger = logging.getLogger('nlu')

def substitute_ner_converter_cols(c, cols, nlu_identifier):
    """
    Fetched fields are:
    - entities@<storage_ref>_results
    - entities@<storage_ref>_<metadata>
        - entities@<storage_ref>_entity
        - entities@<storage_ref>_confidence
    """
    new_cols = {}
    new_base_name = 'entities' if nlu_identifier == 'UNIQUE' else f'entities_{nlu_identifier}'
    for col in cols :
        if 'results'     in col     : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : new_cols[col] = f'{new_base_name}_embedding'
        elif 'meta' in col:
            if 'confidence' in col: new_cols[col]= f"{new_base_name}_confidence"
            elif 'entity' in     col: new_cols[col]= f"{new_base_name}_class"
            elif 'chunk' in     col: new_cols[col]= f"{new_base_name}_origin_chunk"
            elif 'sentence' in     col: new_cols[col]= f"{new_base_name}_origin_sentence"
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols

def substitute_ner_dl_cols(c, cols, nlu_identifier):
    """
    Fetched fields are:
    - entities@<storage_ref>_results
    - entities@<storage_ref>_<metadata>
        - entities@<storage_ref>_entity
        - entities@<storage_ref>_confidence
    """
    new_cols = {}
    new_base_name = 'ner_iob' if nlu_identifier=='UNIQUE' else f'ner_iob_{nlu_identifier}'
    for col in cols :
        if 'results'     in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # always empty and irrelevant new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if 'confidence' in col: new_cols[col]= f"{new_base_name}_confidence"
            elif 'word' in     col: continue # is the same as token col, can be omitted
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols

def substitute_doc_assembler_cols(c, cols, is_unique=True):
    """
    Doc assember is always unique
    Fetched fields are:
    - entities@<storage_ref>_results
    - entities@<storage_ref>_<metadata>
        - entities@<storage_ref>_entity
        - entities@<storage_ref>_confidence
    """
    new_cols = {}
    new_base_name = 'document'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col     : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # irrelevant  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types'      in col     : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col: continue # This field is irrelevant, since meta_sentence in document assembler is always 0
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols

def substitute_sentence_detector_dl_cols(c, cols, nlu_identifier=True):
    """
    Sent detector is always unique
    """
    new_cols = {}
    new_base_name = 'sentence' if nlu_identifier=='UNIQUE' else f'sentence_dl'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Sentence never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if 'sentence_sentence' in col: continue # Seems like an irrelevant field, so drop
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols

def substitute_sentence_detector_pragmatic_cols(c, cols, nlu_identifier=True):
    """
    Sent detector is always unique
    """
    new_cols = {}
    new_base_name = 'sentence' if nlu_identifier=='UNIQUE' else f'sentence_pragmatic'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Sentence never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if 'sentence_sentence' in col: continue # Seems like an irrelevant field, so drop
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols

def substitute_tokenizer_cols(c, cols, is_unique=True):
    """
    Tokenizer is always unique
    """
    new_cols = {}
    new_base_name = 'token'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols



def substitute_word_embed_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Word Embeddings. For Word_Embeddings, some name will be infered, and word_embedding_<name> will become the base name schema
    """
    new_cols = {}
    new_base_name = f'word_embedding_{nlu_identifier}'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col     : continue  # new_cols[col] = new_base_name can be omitted for word_embeddings, maps to the origin token, which will be in the tokenizer col anyways
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            elif 'OOV'     in col  : new_cols[col] = f'{new_base_name}_is_OOV'  # maps to which sentence token comes from
            elif 'isWordStart' in col  : new_cols[col] = f'{new_base_name}_is_word_start'  # maps to which sentence token comes from
            elif 'pieceId' in col  : new_cols[col] = f'{new_base_name}_piece_id'  # maps to which sentence token comes from
            elif '_token' in col  :  continue  # Can be omited, is the same as _result, just maps to origin_token
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
        elif '_embeddings' in col     : new_cols[col] = new_base_name # stores the embeds and represents basically the main result

    return new_cols
def substitute_sent_embed_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Word Embeddings. For Word_Embeddings, some name will be infered, and word_embedding_<name> will become the base name schema
    """
    new_cols = {}
    new_base_name = f'sentence_embedding_{nlu_identifier}'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col     : continue  # new_cols[col] = new_base_name can be omitted for word_embeddings, maps to the origin token, which will be in the tokenizer col anyways
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if 'OOV'     in col  : new_cols[col] = f'{new_base_name}_is_OOV'  # maps to which sentence token comes from
            elif 'isWordStart' in col  : new_cols[col] = f'{new_base_name}_is_word_start'  # maps to which sentence token comes from
            elif 'pieceId' in col  : new_cols[col] = f'{new_base_name}_piece_id'  # maps to which sentence token comes from
            elif '_token' in col  :  continue  # Can be omited, is the same as _result, just maps to origin_token
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
        elif '_embeddings' in col     : new_cols[col] = new_base_name # stores the embeds and represents basically the main result
        elif '_sentence' in col and 'meta' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from

    return new_cols
def substitute_chunk_embed_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for chunk Embeddings. For Word_Embeddings, some name will be infered, and chunk_embedding_<name> will become the base name schema
    """
    new_cols = {}
    new_base_name = f'chunk_embedding_{nlu_identifier}'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col     : continue  # new_cols[col] = new_base_name can be omitted for chunk_embeddings, maps to the origin chunk, which will be in the tokenizer col anyways
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            elif 'isWordStart' in col  : new_cols[col] = f'{new_base_name}_is_word_start'  # maps to which sentence token comes from
            elif 'pieceId' in col  : new_cols[col] = f'{new_base_name}_piece_id'  # maps to which sentence token comes from
            elif '_token' in col  :  continue  # Can be omited, is the same as _result, just maps to origin_token
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
        elif '_embeddings' in col     : new_cols[col] = new_base_name # stores the embeds and represents basically the main result

    return new_cols
def substitute_classifier_dl_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Word Embeddings. For Word_Embeddings, some name will be infered, and word_embedding_<name> will become the base name schema
    """
    new_cols = {}
    new_base_name = f'{nlu_identifier}'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'       in col     :   new_cols[col] = new_base_name #can be omitted for chunk_embeddings, maps to the origin chunk, which will be in the tokenizer col anyways
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue #
        elif '_embeddings' in col     : continue #
        elif 'meta' in col:
            old_base_name = f'meta_{c.info.outputs[0]}'
            metadata = col.split(old_base_name)[-1]
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            elif metadata =='confidence'  :  new_cols[col] = f'{new_base_name}_confidence'  # max confidence over all classes
            else :                   new_cols[col] = f'{new_base_name}{metadata}_confidence'  # confidence field
            # else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')

    return new_cols
def substitute_ngram_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Word Embeddings. ngram will be the new base name
    """
    new_cols = {}
    new_base_name = f'ngram'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col]  = f'{new_base_name}'
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue #
        elif '_embeddings' in col     : continue #
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')

    return new_cols
def substitute_labled_dependency_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Labled dependenecy labeled_dependency will become the base name schema
    """
    new_cols = {}
    new_base_name = f'labeled_dependency'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col]  = f'{new_base_name}'
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue #
        elif '_embeddings' in col     : continue #
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')

    return new_cols
def substitute_un_labled_dependency_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Labled dependenecy unlabeled_dependency will become the base name schema
    """
    new_cols = {}
    new_base_name = f'unlabeled_dependency'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col]  = f'{new_base_name}'
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue #
        elif '_embeddings' in col     : continue #
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'
            elif '_head.begin' in col  : new_cols[col] = f'{new_base_name}_head_begin'
            elif 'head.end' in col  : new_cols[col] = f'{new_base_name}_head_end'
            elif 'head' in col  : new_cols[col] = f'{new_base_name}_head'
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')

    return new_cols
def substitute_pos_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Labled dependenecy unlabeled_dependency will become the base name schema
    """
    new_cols = {}
    new_base_name = f'pos'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col]  = f'{new_base_name}'
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue #
        elif '_embeddings' in col     : continue #
        elif 'meta' in col:
            if   '_sentence' in col   : new_cols[col] = f'{new_base_name}_origin_sentence'
            elif '_word' in col       : continue # can be omitted, is jsut the token
            elif 'confidence' in col  : new_cols[col] = f'{new_base_name}_confidence'
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols
def substitute_norm_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for normalized,  <norm> will be new base col name
    """
    new_cols = {}
    new_base_name = f'norm'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col]  = f'{new_base_name}'
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue #
        elif '_embeddings' in col     : continue #
        elif 'meta' in col:
            if   '_sentence' in col   : new_cols[col] = f'{new_base_name}_origin_sentence'
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols
def substitute_doc_norm_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for normalized,  <norm> will be new base col name
    """
    new_cols = {}
    new_base_name = f'doc_norm'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col]  = f'{new_base_name}'
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue #
        elif '_embeddings' in col     : continue #
        elif 'meta' in col:
            if   '_sentence' in col   : new_cols[col] = f'{new_base_name}_origin_sentence'
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols
def substitute_spell_context_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for normalized,  <spell> will be new base col namem
    1 spell checker is assumed per pipe for now
    """
    new_cols = {}
    new_base_name = f'spell' if nlu_identifier=='UNIQUE' else f'spell_dl'
    for col in cols :
        if '_results'    in col       : new_cols[col]  = f'{new_base_name}'
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue #
        elif '_embeddings' in col     : continue #
        elif 'meta' in col:
            if   '_sentence' in col   : new_cols[col] = f'{new_base_name}_origin_sentence'
            if   '_cost' in col   : new_cols[col] = f'{new_base_name}_cost'
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols
def substitute_spell_symm_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for sym,  <spell> will be new base col name
    1 spell checker is assumed per pipe for now
    """
    new_cols = {}
    new_base_name = f'spell' if nlu_identifier=='UNIQUE' else f'spell_sym'
    for col in cols :
        if '_results'    in col       : new_cols[col]  = f'{new_base_name}'
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue #
        elif '_embeddings' in col     : continue #
        elif 'meta' in col:
            if   '_sentence' in col   : new_cols[col] = f'{new_base_name}_origin_sentence'
            if   '_confidence' in col   : new_cols[col] = f'{new_base_name}_confidence'
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols
def substitute_spell_norvig_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for spell,  <spell> will be new base col name
    1 spell checker is assumed per pipe for now
    """
    new_cols = {}
    new_base_name = f'spell' if nlu_identifier=='UNIQUE' else f'spell_norvig'
    for col in cols :
        if '_results'    in col       : new_cols[col]  = f'{new_base_name}'
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue #
        elif '_embeddings' in col     : continue #
        elif 'meta' in col:
            if   '_sentence' in col   : new_cols[col] = f'{new_base_name}_origin_sentence'
            if   '_confidence' in col   : new_cols[col] = f'{new_base_name}_confidence'
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols
def substitute_word_seg_cols(c, cols, nlu_identifier=True):
    """
    Word_seg is always unique
    """
    new_cols = {}
    new_base_name = 'words_seg'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols
def substitute_stem_cols(c, cols, nlu_identifier=True):
    """
    stem is always unique
    """
    new_cols = {}
    new_base_name = 'stem'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols
def substitute_lem_cols(c, cols, nlu_identifier=True):
    """
    stem is always unique
    """
    new_cols = {}
    new_base_name = 'lem'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols
def substitute_stopwords_cols(c, cols, nlu_identifier=True):
    """
    stem is always unique
    """
    new_cols = {}
    new_base_name = 'stopword_less'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols
def substitute_chunk_cols(c, cols, nlu_identifier=True):
    """
    stem is always unique
    """
    new_cols = {}
    new_base_name = 'matched_pos'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols
def substitute_YAKE_cols(c, cols, nlu_identifier=True):
    """
    stem is always unique
    """
    new_cols = {}
    new_base_name = 'keywords'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            if '_score' in col  : new_cols[col] = f'{new_base_name}_confidence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols
def substitute_marian_cols(c, cols, nlu_identifier=True):
    """
    rename cols with base name either <translated> or if not unique <translated_<lang>>
    """
    new_cols = {}
    new_base_name = 'translated' if nlu_identifier=='UNIQUE' else f'translated_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols
def substitute_T5_cols(c, cols, nlu_identifier=True):
    """
    rename cols with base name either <t5> or if not unique <t5_<task>>
    """
    new_cols = {}
    new_base_name = 't5' if nlu_identifier=='UNIQUE' else f't5_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols

def substitute_sentiment_vivk_cols(c, cols, nlu_identifier=True):
    new_cols = {}
    new_base_name = 'sentiment' if nlu_identifier=='UNIQUE' else f'sentiment_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            elif '_confidence' in col  : new_cols[col] = f'{new_base_name}_confidence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols
def substitute_sentiment_dl_cols(c, cols, nlu_identifier=True):
    new_cols = {}
    new_base_name = 'sentiment' if  nlu_identifier=='UNIQUE' else f'sentiment_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            elif '_confidence' in col  : new_cols[col] = f'{new_base_name}_confidence'  # maps to which sentence token comes from
            elif '_negative' in col  : new_cols[col] = f'{new_base_name}_negative'  # maps to which sentence token comes from
            elif '_positive' in col  : new_cols[col] = f'{new_base_name}_positive'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_multi_classifier_dl_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Word Embeddings. For Word_Embeddings, some name will be infered, and word_embedding_<name> will become the base name schema
    """
    new_cols = {}
    new_base_name = f'{nlu_identifier}'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'       in col     :   new_cols[col] = new_base_name #can be omitted for chunk_embeddings, maps to the origin chunk, which will be in the tokenizer col anyways
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue #
        elif '_embeddings' in col     : continue #
        elif 'meta' in col:
            old_base_name = f'meta_{c.info.outputs[0]}'
            metadata = col.split(old_base_name)[-1]
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            elif metadata =='confidence'  :  new_cols[col] = f'{new_base_name}_confidence'  # max confidence over all classes
            else :                   new_cols[col] = f'{new_base_name}{metadata}_confidence'  # confidence field
            # else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')

    return new_cols



def substitute_date_match_cols(c, cols, nlu_identifier=True):
    """
    stem is always unique
    """
    new_cols = {}
    new_base_name = 'matched_date'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols



def substitute_regex_match_cols(c, cols, nlu_identifier=True):
    """
    stem is always unique
    """
    new_cols = {}
    new_base_name = 'matched_regex'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_text_match_cols(c, cols, nlu_identifier=True):
    """
    stem is always unique
    """
    new_cols = {}
    new_base_name = 'matched_text'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols





## Trainable
def substitute_classifier_dl_approach_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Word Embeddings. For Word_Embeddings, some name will be infered, and word_embedding_<name> will become the base name schema
    """
    new_cols = {}
    new_base_name = f'trained_classifier'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'       in col     :   new_cols[col] = new_base_name #can be omitted for chunk_embeddings, maps to the origin chunk, which will be in the tokenizer col anyways
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue #
        elif '_embeddings' in col     : continue #
        elif 'meta' in col:
            old_base_name = f'meta_{c.info.outputs[0]}'
            metadata = col.split(old_base_name)[-1]
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            elif metadata =='confidence'  :  new_cols[col] = f'{new_base_name}_confidence'  # max confidence over all classes
            else :                   new_cols[col] = f'{new_base_name}{metadata}_confidence'  # confidence field
            # else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_sentiment_vivk_approach_cols(c, cols, nlu_identifier=True):
    new_cols = {}
    new_base_name = 'trained_sentiment'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            elif '_confidence' in col  : new_cols[col] = f'{new_base_name}_confidence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_sentiment_dl_approach_cols(c, cols, nlu_identifier=True):
    new_cols = {}
    new_base_name = 'trained_sentiment'
    for col in cols :
        if '_results'    in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # Token never stores Embeddings  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            elif '_confidence' in col  : new_cols[col] = f'{new_base_name}_confidence'  # maps to which sentence token comes from
            elif '_negative' in col  : new_cols[col] = f'{new_base_name}_negative'  # maps to which sentence token comes from
            elif '_positive' in col  : new_cols[col] = f'{new_base_name}_positive'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_multi_classifier_dl_approach_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Word Embeddings. For Word_Embeddings, some name will be infered, and word_embedding_<name> will become the base name schema
    """
    new_cols = {}
    new_base_name = f'trained_multi_classifier'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'       in col     :   new_cols[col] = new_base_name #can be omitted for chunk_embeddings, maps to the origin chunk, which will be in the tokenizer col anyways
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue #
        elif '_embeddings' in col     : continue #
        elif 'meta' in col:
            old_base_name = f'meta_{c.info.outputs[0]}'
            metadata = col.split(old_base_name)[-1]
            if '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            elif metadata =='confidence'  :  new_cols[col] = f'{new_base_name}_confidence'  # max confidence over all classes
            else :                   new_cols[col] = f'{new_base_name}{metadata}_confidence'  # confidence field
            # else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')

    return new_cols



def substitute_ner_dl_approach_cols(c, cols, nlu_identifier):
    """
    Fetched fields are:
    - entities@<storage_ref>_results
    - entities@<storage_ref>_<metadata>
        - entities@<storage_ref>_entity
        - entities@<storage_ref>_confidence
    """
    new_cols = {}
    new_base_name = 'trained_ner_iob'
    for col in cols :
        if 'results'     in col       : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # always empty and irrelevant new_cols[col] = f'{new_base_name}_embedding'
        elif '_types' in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if 'confidence' in col: new_cols[col]= f"{new_base_name}_confidence"
            elif 'word' in     col: continue # is the same as token col, can be omitted
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_pos_approach_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Labled dependenecy unlabeled_dependency will become the base name schema
    """
    new_cols = {}
    new_base_name = f'trained_pos'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col]  = f'{new_base_name}'
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue #
        elif '_embeddings' in col     : continue #
        elif 'meta' in col:
            if   '_sentence' in col   : new_cols[col] = f'{new_base_name}_origin_sentence'
            elif '_word' in col       : continue # can be omitted, is jsut the token
            elif 'confidence' in col  : new_cols[col] = f'{new_base_name}_confidence'
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_doc2chunk_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Doc2chunk
    """
    new_cols = {}
    new_base_name = f'doc2chunk'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col       : new_cols[col]  = f'{new_base_name}'
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types' in col          : continue #
        elif '_embeddings' in col     : continue #
        elif 'meta' in col:
            logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


from sparknlp_jsl.annotator import *
from sparknlp.base import *
from sparknlp_display import *
class VizUtilsHC():
    """Utils for interfacing with the Spark-NLP-Display lib - licensed Viz"""
    HTML_WRAPPER = """<div class="scroll entities" style="overflow-x: auto; border: 1px solid #e6e9ef; border-radius: 0.25rem; padding: 1rem; margin-bottom: 2.5rem; white-space:pre-wrap">{}</div>"""
    @staticmethod
    def infer_viz_licensed(pipe)->str:
        """For a given NLUPipeline with licensed components, infers which visualizations are applicable. """
        # we go in reverse, which makes NER always take lowest priority and NER feeder annotators have higher priority
        for c in pipe.components[::-1]:
            if isinstance(c.model, TypedDependencyParserModel) : return 'dep'
            if isinstance(c.model, (SentenceEntityResolverModel,ChunkEntityResolverModel)) : return 'resolution'
            if isinstance(c.model, (RelationExtractionDLModel,RelationExtractionDLModel)) : return 'relation'
            if isinstance(c.model, (AssertionDLModel,AssertionLogRegModel)) : return 'assert'
            if isinstance(c.model, (NerConverter,NerConverterInternal)) : return 'ner'


    @staticmethod
    def viz_ner(anno_res, pipe,labels = [] ,  viz_colors={},is_databricks_env =False,write_to_streamlit=False):
        """Infer columns required for ner viz and then viz it.
        viz_colors :  set label colors by specifying hex codes , i.e. viz_colors =  {'LOC':'#800080', 'PER':'#77b5fe'}
        labels : only allow these labels to be displayed. (default: [] - all labels will be displayed)
        """
        document_col,entities_col =  VizUtilsHC.infer_ner_dependencies(pipe)
        ner_vis = NerVisualizer()
        if len(viz_colors) > 0 : ner_vis.set_label_colors(viz_colors)
        if write_to_streamlit :
            import streamlit as st
            HTML = ner_vis.display(anno_res,label_col=entities_col,document_col=document_col, labels=labels, return_html=True )
            CSS,HTML = HTML.split('</style>')
            CSS = CSS + '</style>'
            HTML = f'<div> {HTML} '
            st.markdown(CSS, unsafe_allow_html=True)
            # st.markdown(HTML, unsafe_allow_html=True)
            st.markdown(VizUtilsHC.HTML_WRAPPER.format(HTML), unsafe_allow_html=True)

        elif not is_databricks_env:
            ner_vis.display(anno_res,label_col=entities_col,document_col=document_col, labels=labels )
        else :
            return ner_vis.display(anno_res,label_col=entities_col,document_col=document_col, labels=labels,return_html=True )


    @staticmethod
    def infer_ner_dependencies(pipe):
        """Finds entities and doc cols for ner viz"""
        doc_component      = None
        entities_component = None
        for c in pipe.components:
            if isinstance(c.model, (NerConverter,NerConverterInternal)) :        entities_component  = c
            if isinstance(c.model, DocumentAssembler) :   doc_component = c

        document_col     = doc_component.info.outputs[0]
        entities_col = entities_component.info.outputs[0]
        return document_col, entities_col


    @staticmethod
    def viz_dep(anno_res,pipe,is_databricks_env,write_to_streamlit=False):
        """Viz dep result"""
        pos_col,dep_typ_col,dep_untyp_col  = VizUtilsHC.infer_dep_dependencies(pipe)
        dependency_vis = DependencyParserVisualizer()
        if write_to_streamlit :
            import streamlit as st
            SVG = dependency_vis.display(anno_res,pos_col =pos_col,dependency_col =  dep_untyp_col ,dependency_type_col = dep_typ_col,return_html=True)
            # st.markdown(SVG, unsafe_allow_html=True)
            st.markdown(VizUtilsHC.HTML_WRAPPER.format(SVG), unsafe_allow_html=True)
        elif not is_databricks_env:
            dependency_vis.display(anno_res,pos_col =pos_col,dependency_col =  dep_untyp_col ,dependency_type_col = dep_typ_col)
        else:
            return dependency_vis.display(anno_res,pos_col =pos_col,dependency_col =  dep_untyp_col ,dependency_type_col = dep_typ_col,return_html=True)

    @staticmethod
    def infer_dep_dependencies(pipe):
        """Finds entities,pos,dep_typed,dep_untyped and  doc cols for dep viz viz"""
        # doc_component      = None
        pos_component = None
        dep_untyped_component = None
        dep_typed_component = None
        for c in pipe.components:
            if isinstance(c.model, PerceptronModel) :              pos_component  = c
            if isinstance(c.model, TypedDependencyParserModel) :   dep_typed_component  = c
            if isinstance(c.model, DependencyParserModel) :        dep_untyped_component  = c

        pos_col       = pos_component.info.outputs[0]
        dep_typ_col   = dep_typed_component.info.outputs[0]
        dep_untyp_col = dep_untyped_component.info.outputs[0]
        return pos_col,dep_typ_col,dep_untyp_col




    @staticmethod
    def viz_resolution(anno_res,pipe,viz_colors={},is_databricks_env=False,write_to_streamlit=False):
        """Viz dep result. Set label colors by specifying hex codes, i.e. viz_colors={'TREATMENT':'#800080', 'PROBLEM':'#77b5fe'} """
        entities_col,resolution_col,doc_col  = VizUtilsHC.infer_resolution_dependencies(pipe)
        er_vis = EntityResolverVisualizer()
        if len(viz_colors) > 0 : er_vis.set_label_colors(viz_colors)

        if write_to_streamlit :
            import streamlit as st
            HTML = er_vis.display(anno_res,label_col=entities_col, resolution_col = resolution_col,document_col=doc_col,return_html=True)
            CSS,HTML = HTML.split('</style>')
            CSS = CSS + '</style>'
            HTML = f'<div> {HTML} '
            st.markdown(CSS, unsafe_allow_html=True)
            # st.markdown(HTML, unsafe_allow_html=True)
            st.markdown(VizUtilsHC.HTML_WRAPPER.format(HTML), unsafe_allow_html=True)


        elif not is_databricks_env:
            er_vis.display(anno_res,label_col=entities_col, resolution_col = resolution_col,document_col=doc_col)
        else:
            return  er_vis.display(anno_res,label_col=entities_col, resolution_col = resolution_col,document_col=doc_col,return_html=True)


    @staticmethod
    def infer_resolution_dependencies(pipe):
        """Finds entities_col,resolution_col,doc_col cols for resolution viz viz"""
        entities_component,resolution_component,doc_component = None, None, None
        for c in pipe.components:
            if isinstance(c.model, DocumentAssembler) :              doc_component  = c
            if isinstance(c.model, (NerConverter,NerConverterInternal)) :   entities_component  = c
            if isinstance(c.model, (SentenceEntityResolverModel,ChunkEntityResolverModel)) :        resolution_component  = c

        entities_col       = entities_component.info.outputs[0]
        resolution_col   = resolution_component.info.outputs[0]
        doc_col = doc_component.info.outputs[0]
        return entities_col,resolution_col,doc_col

    @staticmethod
    def viz_relation(anno_res,pipe,is_databricks_env,write_to_streamlit=False):
        """Viz relation result. Set label colors by specifying hex codes, i.e. viz_colors={'TREATMENT':'#800080', 'PROBLEM':'#77b5fe'} """
        relation_col,document_col = VizUtilsHC.infer_relation_dependencies(pipe)
        re_vis = RelationExtractionVisualizer()
        if write_to_streamlit :
            import streamlit as st
            HTML = re_vis.display(anno_res,relation_col = relation_col,document_col = document_col, show_relations=True, return_html=True)
            # st.markdown(HTML, unsafe_allow_html=True)
            st.markdown(VizUtilsHC.HTML_WRAPPER.format(HTML), unsafe_allow_html=True)



        if not is_databricks_env:
            re_vis.display(anno_res,relation_col = relation_col,document_col = document_col, show_relations=True)
        else:
            return  re_vis.display(anno_res,relation_col = relation_col,document_col = document_col, show_relations=True, return_html=True)

    @staticmethod
    def infer_relation_dependencies(pipe):
        """Finds relation_col,document_col  cols for relation viz viz"""
        relation_component,doc_component = None, None
        for c in pipe.components:
            if isinstance(c.model, DocumentAssembler) :              doc_component  = c
            if isinstance(c.model, (RelationExtractionDLModel,RelationExtractionModel)) :   relation_component  = c
        relation_col       = relation_component.info.outputs[0]
        document_col = doc_component.info.outputs[0]
        return relation_col,document_col


    @staticmethod
    def viz_assertion(anno_res,pipe,viz_colors={},is_databricks_env=False,write_to_streamlit=False):
        """Viz relation result. Set label colors by specifying hex codes, i.e. viz_colors={'TREATMENT':'#008080', 'problem':'#800080'} """
        entities_col,assertion_col, doc_col = VizUtilsHC.infer_assertion_dependencies(pipe)
        assertion_vis = AssertionVisualizer()
        if len(viz_colors) > 0 : assertion_vis.set_label_colors(viz_colors)
        if write_to_streamlit :
            import streamlit as st
            HTML = assertion_vis.display(anno_res,label_col = entities_col,assertion_col = assertion_col ,document_col = doc_col,return_html=True)
            # st.markdown(HTML, unsafe_allow_html=True)
            CSS,HTML = HTML.split('</style>')
            CSS = CSS + '</style>'
            HTML = f'<div> {HTML} '
            st.markdown(CSS, unsafe_allow_html=True)
            # st.markdown(HTML, unsafe_allow_html=True)
            st.markdown(VizUtilsHC.HTML_WRAPPER.format(HTML), unsafe_allow_html=True)
        elif not is_databricks_env:
            assertion_vis.display(anno_res,label_col = entities_col,assertion_col = assertion_col ,document_col = doc_col)
        else:
            return  assertion_vis.display(anno_res,label_col = entities_col,assertion_col = assertion_col ,document_col = doc_col,return_html=True)

    @staticmethod
    def infer_assertion_dependencies(pipe):
        """Finds relation_col,document_col  cols for relation viz viz"""
        entities_component,assert_component, doc_component = None, None,None
        for c in pipe.components:
            if isinstance(c.model, DocumentAssembler) :              doc_component  = c
            if isinstance(c.model, (AssertionDLModel,AssertionLogRegModel)) :   assert_component  = c
            if isinstance(c.model, (NerConverter,NerConverterInternal)) :   entities_component  = c

        entities_col       = entities_component.info.outputs[0]
        assertion_col       = assert_component.info.outputs[0]
        doc_col = doc_component.info.outputs[0]
        return entities_col,assertion_col, doc_col





d = {'en':{

# DeIdentificationModel  |||| EXPORTED
'en.de_identify':'deidentify_dl',
'en.de_identify.rules':'deid_rules',
'en.de_identify.clinical':'deidentify_enriched_clinical',
'en.de_identify.large':'deidentify_large',
'en.de_identify.rb':'deidentify_rb',
'en.de_identify.rb_no_regex':'deidentify_rb_no_regex',

# ner DEID  |||| EXPORTED
'en.ner.deid.augmented':'ner_deid_augmented',
'en.ner.deid.biobert':'ner_deid_biobert',
'en.ner.deid.enriched':'ner_deid_enriched',
'en.ner.deid.enriched_biobert':'ner_deid_enriched_biobert',
'en.ner.deid.large':'ner_deid_large',
'en.ner.deid.sd':'ner_deid_sd',
'en.ner.deid.sd_large':'ner_deid_sd_large',
'en.ner.deid.synthetic':'ner_deid_synthetic',
'en.ner.deid.dl':'ner_deidentify_dl',
'en.ner.deid.':'nerdl_deid',

# WordEmbeddings ||||||||||||||EXPORTET
'en.':'embeddings_biovec',
'en.':'embeddings_clinical',
'en.':'embeddings_healthcare',
'en.':'embeddings_healthcare_100d',
'en.':'embeddings_icd10_base',
'en.':'embeddings_icdoem',
'en.':'embeddings_icdoem_2ng',


# pos clinical || EXPORTED
'en.pos.clinical':'pos_clinical',
'en.pos.fast':'pos_fast_med',

# RelationExtraction ||| EXPORTED BUT DEPRECRATED?!??!
'en.extract_relation':'re_bodypart_directions',
'en.extract_relation.bodypart.direction':'re_bodypart_directions',
'en.extract_relation.bodypart.problem':'re_bodypart_problem',
'en.extract_relation.bodypart.proceduretest':'re_bodypart_proceduretest',
'en.extract_relation.chemprot':'re_chemprot_clinical',
'en.extract_relation.clinical':'re_clinical',
'en.extract_relation.date_clinical':'re_date_clinical',
'en.extract_relation.drug_drug_interaction':'re_drug_drug_interaction_clinical',
'en.extract_relation.human_phenotype_gene':'re_human_phenotype_gene_clinical',
'en.extract_relation.temporal_events':'re_temporal_events_clinical',
'en.extract_relation.temporal_events.enriched':'re_temporal_events_enriched_clinical',

# RelationExtractionDL |||| EXOIRTED
'en.extract_relation':'redl_bodypart_direction_biobert',
'en.extract_relation.bodypart.direction':'redl_bodypart_direction_biobert',
'en.extract_relation.bodypart.problem':'redl_bodypart_problem_biobert',
'en.extract_relation.bodypart.procedure':'redl_bodypart_procedure_test_biobert',
'en.extract_relation.chemprot':'redl_chemprot_biobert',
'en.extract_relation.clinical':'redl_clinical_biobert',
'en.extract_relation.date':'redl_date_clinical_biobert',
'en.extract_relation.drug_drug_interaction':'redl_drug_drug_interaction_biobert',
'en.extract_relation.humen_phenotype_gene':'redl_human_phenotype_gene_biobert',
'en.extract_relation.temporal_events':'redl_temporal_events_biobert',


#NERDL Models  ||||  EXPORTED!!!
'en.ner.ade.biobert':'ner_ade_biobert',
'en.ner.ade.clinical':'ner_ade_clinical',
'en.ner.ade.clinical_bert':'ner_ade_clinicalbert',
'en.ner.ade.ade_healthcare':'ner_ade_healthcare',
'en.ner.anatomy':'ner_anatomy',
'en.ner.anatomy.biobert':'ner_anatomy_biobert',
'en.ner.anatomy.coarse':'ner_anatomy_coarse',
'en.ner.anatomy.coarse_biobert':'ner_anatomy_coarse_biobert',
'en.ner.aspect_sentiment':'ner_aspect_based_sentiment',
'en.ner.bacterial_species':'ner_bacterial_species',
'en.ner.bionlp':'ner_bionlp',
'en.ner.bionlp.biobert':'ner_bionlp_biobert',
'en.ner.bionlp.biobert.noncontrib':'ner_bionlp_noncontrib', # todo whats noncontrib?
'en.ner.cancer':'ner_cancer_genetics',
'en.ner.cellular':'ner_cellular',
'en.ner.cellular.biobert':'ner_cellular_biobert',
'en.ner.chemicals':'ner_chemicals',
'en.ner.chemprot':'ner_chemprot_biobert',
'en.ner.chemprot.clinical':'ner_chemprot_clinical',
'en.ner.clinical':'ner_clinical',
'en.ner.clinical.biobert':'ner_clinical_biobert',
'en.ner.clinical.icdem':'ner_clinical_icdem',
'en.ner.clinical.large':'ner_clinical_large',
'en.ner.clinical.noncontrib':'ner_clinical_noncontrib',

### NERDL |||||| EXPORTED (PART2)
'en.ner.diseases.':'ner_diseases',
'en.ner.diseases.biobert':'ner_diseases_biobert',
'en.ner.diseases.large':'ner_diseases_large',
'en.ner.drugs':'ner_drugs',
'en.ner.drugsgreedy':'ner_drugs_greedy',
'en.ner.drugs.large':'ner_drugs_large',
'en.ner.events_biobert':'ner_events_biobert',
'en.ner.events_clinical':'ner_events_clinical',
'en.ner.events_healthcre':'ner_events_healthcare',
'en.ner.financial_contract':'ner_financial_contract',
'en.ner.healthcare':'ner_healthcare',
'en.ner.human_phenotype.gene_biobert':'ner_human_phenotype_gene_biobert',
'en.ner.human_phenotype.gene_clinical':'ner_human_phenotype_gene_clinical',
'en.ner.human_phenotype.go_biobert':'ner_human_phenotype_go_biobert',
'en.ner.human_phenotype.go_clinical':'ner_human_phenotype_go_clinical',
'en.ner.jsl':'ner_jsl',
'en.ner.jsl.biobert':'ner_jsl_biobert',
'en.ner.jsl.enriched':'ner_jsl_enriched',
'en.ner.jsl.enriched_biobert':'ner_jsl_enriched_biobert',
'en.ner.measurements':'ner_measurements_clinical',
'en.ner.medmentions':'ner_medmentions_coarse',
'en.ner.posology':'ner_posology',
'en.ner.posology.biobert':'ner_posology_biobert',
'en.ner.posology.greedy':'ner_posology_greedy',
'en.ner.posology.healthcare':'ner_posology_healthcare',
'en.ner.posology.large':'ner_posology_large',
'en.ner.posology.large_biobert':'ner_posology_large_biobert',
'en.ner.posology.small':'ner_posology_small',
'en.ner.radiology':'ner_radiology',
'en.ner.radiology.wip_clinical':'ner_radiology_wip_clinical',
'en.ner.risk_factors':'ner_risk_factors',
'en.ner.risk_factors.biobert':'ner_risk_factors_biobert',
'en.ner.i2b2':'nerdl_i2b2',
'en.ner.tumour':'nerdl_tumour_demo',
'en.ner.jsl.wip.clinical':'jsl_ner_wip_clinical',
'en.ner.jsl.wip.clinical.greedy':'jsl_ner_wip_greedy_clinical',
'en.ner.jsl.wip.clinical.modifier':'jsl_ner_wip_modifier_clinical',
'en.ner.jsl.wip.clinical.rd':'jsl_rd_ner_wip_greedy_clinical',




# SentenceBertEmbedding |||||||||||||| EXPORTERD
'en.':'sent_biobert_base_uncased_mednli',
'en.':'sent_bluebert_base_uncased_mednli',
'en.resolve_sentence.mli.bluebuert':'sbluebert_base_uncased_mli',


#ASSERT (so smul) AssertionDLModel ||| EXPORTED
'en.assert':'assertion_dl', # clinicalz
'en.assert.biobert':'assertion_dl_biobert', # biobert_pubmed_base_cased
'en.assert.healthcare':'assertion_dl_healthcare', # embeddings_healthcare100
'en.assert.large':'assertion_dl_large', # clinical



#BIOBERT embeddings  |||||||||||| EXPORTED
'en.embed.biobert.clinical':'biobert_clinical_base_cased', # Todo crash NO POOLING LAYER
'en.embed.biobert.discharge':'biobert_discharge_base_cased',
'en.embed.biobert.pmc':'biobert_pmc_base_cased',
'en.embed.biobert.pubmed':'biobert_pubmed_base_cased', # Todo crash NO POOLING LAYER
'en.embed.biobert.pubmed_large':'biobert_pubmed_large_cased',
'en.embed.biobert.pubmed_pmc':'biobert_pubmed_pmc_base_cased',


# sentence Entity resolvers ||||||||EX{PRTED
'en.':'sbiobert_base_cased_mli',
'en.':'sbiobertresolve_cpt',
'en.':'sbiobertresolve_cpt_augmented',
'en.':'sbiobertresolve_cpt_procedures_augmented',
'en.':'sbiobertresolve_hcc_augmented',
'en.':'sbiobertresolve_icd10cm',
'en.':'sbiobertresolve_icd10cm_augmented',
'en.':'sbiobertresolve_icd10cm_augmented_billable_hcc',
'en.':'sbiobertresolve_icd10pcs',
'en.':'sbiobertresolve_icdo',
'en.':'sbiobertresolve_rxcui',
'en.':'sbiobertresolve_rxnorm',
'en.':'sbiobertresolve_snomed_auxConcepts',
'en.':'sbiobertresolve_snomed_auxConcepts_int',
'en.':'sbiobertresolve_snomed_findings',
'en.':'sbiobertresolve_snomed_findings_int',
'en.':'sbluebert_base_uncased_mli',





# ClassifierDL Models       +++++++++}|||||||||||||||||||||||||||||||||||||| EXPORTED `
'en.classify.icd10.clinical':'classifier_icd10cm_hcc_clinical', # WHCIH CLASS? # TODO NOT LAODING
'en.classify.icd10.healthcare':'classifier_icd10cm_hcc_healthcare', # TODO NOT LOADING CORRECt
'en.classify.ade.biobert':'classifierdl_ade_biobert',
'en.classify.ade.clinical':'classifierdl_ade_clinicalbert',
'en.classify.ade.conversational':'classifierdl_ade_conversational_biobert',
'en.classify.gender.biobert':'classifierdl_gender_biobert',
'en.classify.gender.sbert':'classifierdl_gender_sbert', # ok!
'en.classify.pico':'classifierdl_pico_biobert',
'en.classify.icd10.use':'useclassifier_icd10cm_hcc',

    # Chunk resolve ||||||||| EXPORTED
    'en.resolve_chunk.athena_conditions':'chunkresolve_athena_conditions_healthcare',
    'en.resolve_chunk.cpt_clinical':'chunkresolve_cpt_clinical',
    'en.resolve_chunk.cpt_icdoem':'chunkresolve_cpt_icdoem',
    'en.resolve_chunk.icd10cm.clinical':'chunkresolve_icd10cm_clinical',
    'en.resolve_chunk.icd10cm.diseases_clinical':'chunkresolve_icd10cm_diseases_clinical',
    'en.resolve_chunk.icd10cm.hcc_clinical':'chunkresolve_icd10cm_hcc_clinical',
    'en.resolve_chunk.icd10cm.hcc_healthcare':'chunkresolve_icd10cm_hcc_healthcare',
    'en.resolve_chunk.icd10cm.icdoem,':'chunkresolve_icd10cm_icdoem',
    'en.resolve_chunk.icd10cm.icdoem_2ng':'chunkresolve_icd10cm_icdoem_2ng',
    'en.resolve_chunk.icd10cm.injuries':'chunkresolve_icd10cm_injuries_clinical',
    'en.resolve_chunk.icd10cm.musculoskeletal':'chunkresolve_icd10cm_musculoskeletal_clinical',
    'en.resolve_chunk.icd10cm.neoplasms':'chunkresolve_icd10cm_neoplasms_clinical',
    'en.resolve_chunk.icd10cm.poison':'chunkresolve_icd10cm_poison_ext_clinical',
    'en.resolve_chunk.icd10cm.puerile':'chunkresolve_icd10cm_puerile_clinical',
    'en.resolve_chunk.icd10cpt.icdoem':'chunkresolve_icd10cpt_icdoem_2ng',
    'en.resolve_chunk.icd10pcs.clinical':'chunkresolve_icd10pcs_clinical',
    'en.resolve_chunk.icd10pcs.icdoem':'chunkresolve_icd10pcs_icdoem',
    'en.resolve_chunk.icd10pcs.icdoem_2ng':'chunkresolve_icd10pcs_icdoem_2ng',
    'en.resolve_chunk.icdo.clinical':'chunkresolve_icdo_clinical',
    'en.resolve_chunk.icdo.icdoem':'chunkresolve_icdo_icdoem',
    'en.resolve_chunk.loinc':'chunkresolve_loinc_clinical',
    'en.resolve_chunk.rxnorm.cd':'chunkresolve_rxnorm_cd_clinical',
    'en.resolve_chunk.rxnorm.in':'chunkresolve_rxnorm_in_clinical',
    'en.resolve_chunk.rxnorm.in_healthcare':'chunkresolve_rxnorm_in_healthcare',
    'en.resolve_chunk.rxnorm.sbd':'chunkresolve_rxnorm_sbd_clinical',
    'en.resolve_chunk.rxnorm.scd':'chunkresolve_rxnorm_scd_clinical',
    'en.resolve_chunk.rxnorm.scdc':'chunkresolve_rxnorm_scdc_clinical',
    'en.resolve_chunk.rxnorm.scdc_healthcare':'chunkresolve_rxnorm_scdc_healthcare',
    'en.resolve_chunk.rxnorm.xsmall.clinical':'chunkresolve_rxnorm_xsmall_clinical',
    'en.resolve_chunk.snomed.findings':'chunkresolve_snomed_findings_clinical',

},
    'de' :{
        'de.embed' :'w2v_cc_300d',
        'de.embed.w2v' :'w2v_cc_300d',
        'de.resolve_chunk' :'chunkresolve_ICD10GM',
        'de.resolve_chunk.icd10gm' :'chunkresolve_ICD10GM',
        'de.resolve_chunk.icd10gm.2021' :'chunkresolve_ICD10GM_2021',
        'de.med_ner.legal' :'ner_legal',
        'de.med_ner' :'ner_healthcare',
        'de.med_ner.healthcare' :'ner_healthcare',
        'de.med_ner.healthcare_slim' :'ner_healthcare_slim',
        'de.med_ner.traffic' :'ner_traffic',
    }
}


#===========================+EXPORT END ==================================


# relation extractor pipe posology
'en.':'recognize_entities_posology',

# AssertionLogRegModel
'en.':'assertion_ml',
# SentenceEntityResolverModel biobert
# These all hava analogus models with s prefixed to them. Deprecated?
# Sentence Entity Resolvers Deprecated models
# 'en.':'-biobertresolve_cpt', #  TODO Crash?
# 'en.':'-biobertresolve_icd10cm', # TODO crash
# 'en.':'-biobertresolve_icd10pcs', # todo crash
# 'en.':'-biobertresolve_icdo', # todo crash
# 'en.':'-biobertresolve_loinc', # todo crash
# 'en.':'-biobertresolve_snomed_findings',
# 'en.':'-biobertresolve_snomed_findings_int',

# 'en.':'clinical_analysis', # TODO conver t to spark 3
'en.':'clinical_deidentification',
'en.':'clinical_ner_assertion',
'en.':'resolve_icd10',








# NO HUB NO CARE!?
# resolve cpt (SENTENCE OR CHUNK??) or pipe?
'en.':'resolve_cpt_cl_em',
'en.':'resolve_cpt_icdoem',



'en.':'resolve_icd10cm_cl_em',
'en.':'resolve_icd10cm_icdem',
'en.':'resolve_icd10cm_icdoem',
'en.':'resolve_icd10pcs_cl_em',
'en.':'resolve_icdo_icdoem',
'en.':'resolve_rxnorm_clinical_l1',
'en.':'resolve_rxnorm_clinical_l2',
'en.':'resolve_rxnorm_healthcare_l1',
'en.':'resolve_rxnorm_healthcare_l2',
'en.':'resolve_rxnorm_l1_idx_icdoem_2ng',
'en.':'resolve_rxnorm_l1_ovrlrc_icdoem_2ng',
'en.':'resolve_rxnorm_l1_tfidf_icdoem_2ng',
'en.':'resolve_rxnorm_l2_icdoem_2ng',
# 'en.':'-biobertresolve_icd10cm_hcc',  # TODO crash
# 'en.':'-biobertresolve_rxnorm_bdcd',

# ensemble resolve # TODO class?
'en.':'ensembleresolve_icd10cm_clinical',
'en.':'ensembleresolve_rxnorm_small_clinical',
'en.':'ensembleresolve_snomed_small_clinical',


# entity resolve
'en.':'entity_resolver_icd10',


# ? (pipes?) or disambigioators  NOT IN MODEL HUB, NOT CARE!
'en.':'people_disambiguator',
# pipe? NOT IN MODEL HUB, NOT CARE!
'en.':'ppl_posology_rxnorm',


# resolve snomed (SENTENCE OR CHUNK??) or pipe?
# 'en.resolve_?.snomed':'resolve_snomed_clinical_l1',
# 'en.resolve_?.snomed.clinical.l1':'resolve_snomed_clinical_l1',
# 'en.resolve_?.snomed.clinical.l2':'resolve_snomed_clinical_l2',
# 'en.resolve_?.snomed.l1.idx':'resolve_snomed_l1_idx_icdoem_2ng',
# 'en.resolve_?.snomed.l1.ovrlrc':'resolve_snomed_l1_ovrlrc_icdoem_2ng',
# 'en.resolve_?.snomed.l1.tfidf':'resolve_snomed_l1_tfidf_icdoem_2ng',
# 'en.resolve_?.snomed.l2.icdoem':'resolve_snomed_l2_icdoem_2ng',
#

########################### END OF NONE MODEL HBUBSZ

# text matchcher old deprecated/ client project
# 'en.':'textmatch_cpt_token',
# 'en.':'textmatch_cpt_token_n2c1',
# 'en.':'textmatch_icd10cm',
# 'en.':'textmatch_icd10cm_uncased',
# 'en.':'textmatch_icdo',
# 'en.':'textmatch_icdo_ner',
# 'en.':'textmatch_icdo_ner_n2c4',
#


}




# ner CRF (DEPARACTD)
# 'en.':'ner_crf',
# ner CRF (DEPRACTATED)
# 'en.':'nercrf_deid',z
# TODO JSL STUFF IS ARE UNDOCUMENTED CLIENT PROEJCTS! exclude for now!
# Jsl assert
'en.':'jsl_assertion_wip',
'en.':'jsl_assertion_wip_large',
# JSL embeds
'en.':'jsl_bert_pubmed_cased',








import nlu
default_text = ('What is love?', 'Baby dont hurt me')
nlu.load('ner').viz_streamlit(default_text, visualizers=['embedding_sentence'])
nlu.load('ner').viz_streamlit(default_text, visualizers=['embedding_words'])



import sparknlp
from sparknlp.annotator import TextMatcherModel
class TextMatcher:
    @staticmethod
    def get_default_model():
        return sparknlp.annotator.TextMatcher() \
            .setInputCols("sentence","token") \
            .setOutputCol("matched_entity")


    @staticmethod
    def get_pretrained_model(name, language):
        return sparknlp.annotator.TextMatcherModel.pretrained(name,language) \
            .setInputCols("sentence","token") \
            .setOutputCol("matched_entity") \





import nlu
# TODO
# display_vector_simmilarity()
# display_entity_simmilarity()

# TOO unremovable footers?
# TODO sentence/doc similarity
# TODO auto-detect GPU? If TF avaiable, try to check if TF gpu avaiable?
# TODO XIBIT, YO I HEARD YOU LIKE EMBEDDINGS, SO we made a TOOl to embed you embeddings into 1-D, 2-D, and 3-D
# TODO INSTALL INSTRUCTIONS
# TODO 3D Similarity Matrix/CUBE/HYPER-SPHERE??!?!??!?!
# TODO SLACK LINK 4 COMMUNITY/ NLP DISCUSS FORUM
# TODO default sentiment classifier and no NER?
# TODO SENTENCE SIMILARITY
# TODO trigger ALL model dropdowns via show_model_select
# Todo model_selection for each app? Or just via audo infer?
# Todo extra classifiers, sentiment?
# todo remove XX for translators
# Todo code snippet on left side is wrong and for viz_word_embed aswell
# TODO always display LOGO
# TODO not all pipes listed in pipe param info
# TODO POS SELECTOR # TODO POS MODEL SELECT?
# AR NER BUGGY !
# LANGUAGE DETECT TOGGLE!
# NER.onto.lg BUGGY!
nlu.enable_streamlit_caching() # Optional caching the models, recommended
# Peter is from Berlin and where he loves visting the Charlottenburg Palace
#
data = 'Peter is from Berlin and where he loves visting the Charlottenburg Palace'
# # NER -> onto.sm,  NER-BERT-MINI -> BIGGER NER -> French NER ->
# 1. default text,
# 2. "Hello from john Snow Labs"
# 3. Show Bert
# 4. Wikipedia Math + lang https://fr.wikipedia.org/wiki/Histoire_des_math%C3%A9matiques
# 5 Some extra lang  ru.ner,tr.ner , fr.ner ,  de.ner, it.ner
nlu.load('ner').viz_streamlit(data)

# nlu.load('ner').viz_streamlit_ner(data)
#
# # EMBEDS
# 1. bert elmo glove
# + SIM
# 2.albert electra xlnet
# - sim
# 3. biobert / Covid /  small bert
#
# 2 or 3 metrics
# I love peanutbutter on my bread /// AND jelly!
# You can use 100's of Word embeddings to compare your texts
# I love love looove peanut butter for breakfast !
# and jelly <3
# and calculate similarities for!


# nlu.load('bert').viz_streamlit_word_similarity(['I love love loooove NLU! <3','I also love love looove  Streamlit! <3'], num_cols=3)
#
#
# # TOKEN
# lema, stem, pos, spell ,??
# data = 'You can play around with any of the 1000+ models in this visualization and view features generated'
# nlu.load('spell stem lemma pos').viz_streamlit_token('I liek pentut buttr and jelly !')
#
#
# # DEP TREEE
# nlu.load('dep.typed').viz_streamlit_dep_tree('POS tags define a grammatical label for each token and the Dependency Tree classifies Relations between the tokens')
# # CLASSIFIER
# sentiment, sarcasm, emotion, question, spam,
#
# nlu.load('sentiment').viz_streamlit_classes(['I love NLU and Streamlit!','I love buggy software', 'Sign up now get a chance to win 1000$ !', 'I am afraid of Snakes','Unicorns have been sighted on Mars!','Where is the next bus stop?'],sub_title=sh)

# FULL UI TODO
# # TODO automatically show all embeds and NER?
# ner_models = nlu.get_components('ner',include_pipes=True)
# default_text = "Angela Merkel from Germany and Donald Trump from America dont share many opinions"
# # nlu.load('ner').viz_streamlit(default_text, visualizers=['similarity'])
# #

# nlu.load('ner').viz_streamlit(visualizers=['similarity'])
# nlu.load('ner').viz_streamlit(similarity_texts=['I love love love  NLU !', 'I also love love love LOVE Streamlit'])
# nlu.load('ner').viz_streamlit(visualizers=['similarity','ner', 'dependency_tree'], num_similarity_cols=3)


# TODO dependencies sparknlp_display, sklearn, plotly



# 1. NER :  Peter loves visiting Charlottenburg Palace in Berlin
#   1.1NER -BERT
# 2 DEP:  billy likes to swim on sunny days
# 3 Token :  stem/pos/lemma/ There are 1000s of models to play with //
# 4. Class : Sentiment/Emotion :  I love sunny days, I hate rainy days
# 5. Embeds - Elmo, Bert, Glove, Albert, Electra, Xlnet
#

# nlu.load('ner').viz_streamlit(['I love NLU and Streamlit!','I hate buggy software'])



from sparknlp.annotator import *

class T5:
    @staticmethod
    def nlu_ref_to_task(task):
        nlu_to_task = {
            'check_grammar': 'cola sentence: ',
            'summarize': 'summarize: ',
            'sentiment': 'sst2 sentence:  ',
            'answer_question': 'question:  ',

        }
        if task in nlu_to_task.keys() :return nlu_to_task[task]
        else : return task
    @staticmethod
    def get_default_model():
        return T5Transformer.pretrained() \
        .setInputCols("document") \
        .setOutputCol("T5")

    @staticmethod
    def get_pretrained_model(name, language,bucket=None):
        return T5Transformer.pretrained(name, language,bucket) \
            .setInputCols("document") \
            .setOutputCol("T5")

    @staticmethod
    # Get T5 with task pre-configured
    # Sets task either to somthing known or whatever what detected
    def get_preconfigured_model(name, language,task):

        return T5Transformer.pretrained(name, language) \
            .setInputCols("document") \
            .setOutputCol("T5")\
            .setTask(T5.nlu_ref_to_task(task))








import nlu
from nlu.pipe.pipe_components import SparkNLUComponent


class Matcher(SparkNLUComponent):
    def __init__(self, annotator_class='date_matcher', language='en', component_type='matcher', get_default=False,
                 nlp_ref='', model=None, nlu_ref='', dataset='', is_licensed=False, loaded_from_pretrained_pipe=False):

        if 'date' in nlp_ref or 'date' in nlu_ref:
            annotator_class = 'date_matcher'
        elif 'regex' in nlp_ref or 'regex' in nlu_ref:
            annotator_class = 'regex_matcher'
        elif 'context' in nlu_ref:
            annotator_class = 'context_parser'
        elif 'text' in nlp_ref or 'text' in nlu_ref:
            annotator_class = 'text_matcher'
        elif '_matcher' not in annotator_class:
            annotator_class = annotator_class + '_matcher'
        if model != None:
            self.model = model
        else:
            if 'context' in annotator_class:
                from nlu.components.matchers.context_parser.context_parser import ContextParser
                is_licensed = True
                if get_default:
                    self.model = ContextParser.get_default_model()
                else:
                    self.model = ContextParser.get_default_model()

            elif 'text' in annotator_class:
                from nlu import TextMatcher
                if get_default or nlp_ref =='text_matcher':
                    self.model = TextMatcher.get_default_model()
                else:
                    self.model = TextMatcher.get_pretrained_model(nlp_ref, language)
            elif 'date' in annotator_class:
                from nlu import DateMatcher
                from nlu.components.matchers.date_matcher.date_matcher import DateMatcher as DateM

                if get_default: self.model = DateM.get_default_model()
                else: self.model = DateM.get_default_model()
            elif 'regex' in annotator_class:
                from nlu import RegexMatcher
                if get_default:
                    self.model = RegexMatcher.get_default_model()
                else:
                    self.model = RegexMatcher.get_pretrained_model(nlu_ref, language)

        SparkNLUComponent.__init__(self, annotator_class, component_type, nlu_ref, nlp_ref, language,loaded_from_pretrained_pipe , is_licensed)



from sparknlp.annotator import *

class ValidateVizPipe():
    """Verify for various visualizatins of dependenciesa re satisfied by pipe """

    @staticmethod
    def viz_tree_satisfied(pipe):
        un_typed_dep  = False
        typed_dep  = False
        for c in pipe.components :
            if isinstance(c.model, (DependencyParserModel)):      un_typed_dep = True
            if isinstance(c.model, (TypedDependencyParserModel)): typed_dep    = True
        return un_typed_dep and typed_dep

    @staticmethod
    def viz_NER_satisfied(pipe):
        NER  = False
        NER_CONV  = False
        for c in pipe.components :
            if isinstance(c.model, (NerDLModel)):      NER = True
            if isinstance(c.model, (NerConverter)):    NER_CONV    = True
        return NER and NER_CONV




import unittest
import nlu
import pandas as pd
import numpy as np


class TestComponentInfo(unittest.TestCase):
    def test_list_all_names(self):
        a = nlu.AllComponentsInfo()
        a.list_all_components()
        a.DEBUG_list_all_components()

    def test_print_all_default_components_as_markdown(self):
        d = nlu.Spellbook.component_alias_references
        f = lambda x : x[0]
        g = lambda x : x[1]
        print(pd.DataFrame(zip(d.keys(),map(f,d.values()),map(g,d.values()))).to_markdown(index=False))

    def test_print_all_models_as_markdown(self):
        rows = []
        for lang in nlu.Spellbook.pretrained_models_references.keys():
            for nlu_reference in nlu.Spellbook.pretrained_models_references[lang].keys():
                rows.append((lang,nlu_reference,nlu.Spellbook.pretrained_models_references[lang][nlu_reference]))
        
        print(pd.DataFrame(rows).to_markdown(index=False))
        
    def test_print_all_pipes_as_markdown(self):
        rows = []
        for lang in nlu.Spellbook.pretrained_pipe_references.keys():
            for nlu_reference in nlu.Spellbook.pretrained_pipe_references[lang].keys():
                rows.append((lang,nlu_reference,nlu.Spellbook.pretrained_pipe_references[lang][nlu_reference]))

        print(pd.DataFrame(rows).to_markdown(index=False))


    def test_print_all_HC_models_as_markdown(self):
        rows = []
        for lang in nlu.Spellbook.pretrained_healthcare_model_references.keys():
            for nlu_reference in nlu.Spellbook.pretrained_healthcare_model_references[lang].keys():
                rows.append((lang,nlu_reference,nlu.Spellbook.pretrained_healthcare_model_references[lang][nlu_reference]))

        print(pd.DataFrame(rows).to_markdown(index=False))

    def test_print_all_HC_pipes_as_markdown(self):
        rows = []
        for lang in nlu.Spellbook.pretrained_healthcare_pipe_references.keys():
            for nlu_reference in nlu.Spellbook.pretrained_healthcare_pipe_references[lang].keys():
                rows.append((lang,nlu_reference,nlu.Spellbook.pretrained_healthcare_pipe_references[lang][nlu_reference]))

        print(pd.DataFrame(rows).to_markdown(index=False))


    def test_get_count_for_every_component_type(self):
        component_counts = {}
        for lang in nlu.Spellbook.pretrained_models_references:
            for nlu_ref, nlp_ref in nlu.Spellbook.pretrained_models_references[lang].items():
                c_type = nlu_ref.split('.')
                
                if len(c_type) <2 : continue
                c_type = c_type[1]
                
                if c_type not in component_counts : component_counts[c_type] = []
                if (lang+nlp_ref) not in component_counts[c_type] : component_counts[c_type].append(lang+nlp_ref)
                    
        
        for c_type, components in component_counts.items(): 
            print(c_type, len(components))
  

    def test_get_language_count_for_every_component_type(self):
        component_counts = {}
        for lang in nlu.Spellbook.pretrained_models_references:
            for nlu_ref, nlp_ref in nlu.Spellbook.pretrained_models_references[lang].items():
                c_type = nlu_ref.split('.')

                if len(c_type) <2 : continue
                c_type = c_type[1]

                if c_type not in component_counts : component_counts[c_type] = []
                if (lang+nlp_ref) not in component_counts[c_type] : component_counts[c_type].append(lang+nlp_ref)


        for c_type, components in component_counts.items():
            print(c_type, len(components))

    
    def test_get_count_of_unique_spark_nlp_references(self):
        spark_nlp_references = []

        # # get refs from default        
        # for key, value in nlu.NameSpace.component_alias_references.items():
        #     if value[0] not in spark_nlp_references : spark_nlp_references.append(value[0])

        # get refs from pipes
        pipes = []
        for lang in nlu.Spellbook.pretrained_pipe_references.keys() :
            for key,value in nlu.Spellbook.pretrained_pipe_references[lang].items():
                if lang+value not in spark_nlp_references : 
                    pipes.append(value)
                    spark_nlp_references.append(lang+value)

        models = []
        # get refs from models
        for lang in nlu.Spellbook.pretrained_models_references.keys() :
            for key,value in nlu.Spellbook.pretrained_models_references[lang].items():
                if lang+value not in spark_nlp_references :
                    models.append(value)
                    spark_nlp_references.append(lang+value)
        print("Num overall references", len(spark_nlp_references))
        print("Num Model references", len(models))
        print("Num Pipe references", len(pipes))
        
        
        print(spark_nlp_references)
        
    def test_get_count_of_unique_languages(self):
        print('num languages in NLU : ', len(nlu.all_components_info.all_languages))
    
    
    def test_print_pipe_info(self):
        pipe = nlu.load('sentiment')
        
        pipe.generate_class_metadata_table()
        
    def test_print_all_components_for_lang(self):
        # Test printing of all components for one specific language
        nlu.print_components( 'de')


    def test_print_all_component_types(self):
        # Test printing of all components types
        nlu.print_component_types()
        

        
    def test_print_all_components_for_action_in_lang(self):
        # Test printing of all components for one specific action and language
        nlu.print_components( lang='en', action='classify')

    def test_print_all_components_for_action(self):
        # Test printing of all components for one specific type
        nlu.print_components(action='lemma')


    def test_print_all_components(self):
        nlu.print_components()
    def test_print_all_trainable_components(self):
        nlu.print_trainable_components()


    
if __name__ == '__main__':
    TestComponentInfo().test_entities_config()



import nlu
import streamlit as st
from sparknlp.annotator import NerConverter,DependencyParserModel
from typing import List, Tuple, Optional, Dict, Union
import streamlit as st
from nlu.utils.modelhub.modelhub_utils import ModelHubUtils
import numpy as np
import pandas as pd
from sparknlp.annotator import *
from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
from nlu.pipe.viz.streamlit_viz.gen_streamlit_code import get_code_for_viz
def _set_block_container_style(
        max_width: int = 1200,
        max_width_100_percent: bool = True,
        set_colors : bool = False ,
        COLOR : str = 'blue',
        BACKGROUND_COLOR : str = 'white',
):
    # .reportview-container .main .block-container
    if max_width_100_percent:
        max_width_str = f"max-width: 100%;"
    else:
        max_width_str = f"max-width: {max_width}px;"
    if set_colors:
        st.markdown(
            f"""
    <style>
        .reportview-container .main .block-container{{
            {max_width_str}

        }}
        
     .element-container div:nth-child(1) {{
              margin: auto;

        }}
        
        .reportview-container .main {{
            color: {COLOR};
            background-color: {BACKGROUND_COLOR};
        }}
        
    </style>
    """,
            unsafe_allow_html=True,
        )
    else :
        st.markdown(
            f"""
<style>
    .reportview-container .main .block-container{{
        {max_width_str}
          margin: auto;

    }}
     .element-container div:nth-child(1) {{
              margin: auto;

        }}
</style>
""",
            unsafe_allow_html=True,
        )
_set_block_container_style()

def find_all_embed_components(p):
    """Find ALL component in pipe"""
    cs = []
    for c in p.components :
        if 'embed' in c.info.outputs[0] : cs.append(c)
    if len(cs) == 0 : st.warning("No Embed model in pipe")
    return cs

def get_embed_vetor_information(embed_component,embed_mat,name=''):
    name = StreamlitUtilsOS.extract_name(embed_component) if name =='' else name
    if name =='': name = 'See modelshub for more details'
    return {"Vector Dimension ":embed_mat.shape[1],
              "Num Vectors":embed_mat.shape[0] + embed_mat.shape[0],
              'Vector Name':name}

def display_word_similarity(
        pipe, #nlu pipe
        default_texts: Tuple[str, str] = ("Donald Trump likes to party!", "Angela Merkel likes to party!"),
        threshold: float = 0.5,
        title: Optional[str] = "Vectors & Scalar Similarity & Vector Similarity & Embedding Visualizations  ",
        write_raw_pandas : bool = False ,
        display_embed_information:bool = True,
        similarity_matrix = True,
        show_algo_select : bool = True,
        dist_metrics:List[str]  =('cosine'),
        set_wide_layout_CSS:bool=True,
        generate_code_sample:bool = False,
        key:str = "RANDOM",
        num_cols:int=2,
        display_scalar_similarities : bool = False ,
):

    """We visualize the following cases :
    1. Simmilarity between 2 words - > sim (word_emb1, word_emb2)
    2. Simmilarity between 2 sentences -> let weTW stand word word_emb of token T and sentence S
        2.1. Raw token level with merged embeddings -> sim([we11,we21,weT1], [we12,we22,weT2])
        2.2  Autogenerate sentemb, basically does 2.1 in the Spark NLP backend
        2.3 Already using sentence_embedder model -> sim(se1,se2)
    3. Simmilarity between token and sentence -> sim([we11,w21,wT1], se2)
    4. Mirrored 3
     """
    from nlu.discovery import Discoverer
    # https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics.pairwise
    # TODO try/except sklearn import
    from sklearn.metrics.pairwise import distance_metrics
    if set_wide_layout_CSS : _set_block_container_style()
    if title:st.header(title)
    dist_metric_algos =distance_metrics()
    dist_algos = list(dist_metric_algos.keys())
    # TODO NORMALIZE DISTANCES TO [0,1] for non cosine
    if 'haversine'   in dist_algos    : dist_algos.remove('haversine') # not applicable in >2D
    if 'precomputed' in dist_algos  : dist_algos.remove('precomputed') # Not a dist
    cols = st.beta_columns(2)
    text1 = cols[0].text_input("Text or word1",default_texts[0],key = key)
    text2 = cols[1].text_input("Text or word2",default_texts[1], key=key) if len(default_texts) >1  else cols[1].text_input("Text or word2",'Please enter second string',key = key)
    exp = st.beta_expander("Select additional Embedding Models and distance metric to compare ")
    e_coms = StreamlitUtilsOS.find_all_embed_components(pipe)
    embed_algos_to_load = []
    embed_pipes = [pipe]
    if show_algo_select :
        # emb_components_usable = Discoverer.get_components('embed')
        emb_components_usable = [e for e in Discoverer.get_components('embed',True, include_aliases=True) if 'chunk' not in e and 'sentence' not in e]
        loaded_embed_nlu_refs = []
        for c in e_coms :
            if not  hasattr(c.info,'nlu_ref'): continue
            r = c.info.nlu_ref
            if 'en.' not in r and 'embed.' not  in r : loaded_embed_nlu_refs.append('en.embed.' + r)
            if 'en.'  in r and 'embed.' not  in r :
                r = r.split('en.')[0]
                loaded_embed_nlu_refs.append('en.embed.' + r)

        embed_algo_selection = exp.multiselect("Click to pick additional Embedding Algorithm",options=emb_components_usable,default=loaded_embed_nlu_refs,key = key)
        dist_algo_selection = exp.multiselect("Click to pick additional Distance Metric", options=dist_algos, default=dist_metrics, key = key)
        embed_algos_to_load = list(set(embed_algo_selection) - set(loaded_embed_nlu_refs))

    else :dist_algo_selection = dist_metrics
    for embedder in embed_algos_to_load: embed_pipes.append(nlu.load(embedder))
    similarity_metrics = {}
    embed_vector_info = {}
    cols_full = True
    col_index=0
    for p in embed_pipes :
        data1 = p.predict(text1,output_level='token')
        data2 = p.predict(text2,output_level='token')
        e_coms = StreamlitUtilsOS.find_all_embed_components(p)
        modelhub_links = [ModelHubUtils.get_url_by_nlu_refrence(c.info.nlu_ref) if hasattr(c.info,'nlu_ref') else ModelHubUtils.get_url_by_nlu_refrence('') for c in e_coms]
        e_cols = StreamlitUtilsOS.get_embed_cols(p)
        for num_emb,e_col in enumerate(e_cols):
            if col_index == num_cols-1 :cols_full=True
            if cols_full :
                cols = st.beta_columns(num_cols)
                col_index = 0
                cols_full = False
            else:col_index+=1
            tok1 = data1['token']
            tok2 = data2['token']
            emb1 = data1[e_col]
            emb2 = data2[e_col]
            embed_mat1 = np.array([x for x in emb1])
            embed_mat2 = np.array([x for x in emb2])
            e_name = e_col.split('word_embedding_')[-1]
            embed_vector_info[e_name]= {"Vector Dimension ":embed_mat1.shape[1],"Num Vectors":embed_mat1.shape[0] + embed_mat1.shape[0],'Modelhub info': modelhub_links[num_emb]}
            for dist_algo in dist_algo_selection:
                # scalar_similarities[e_col][dist_algo]={}
                sim_score = dist_metric_algos[dist_algo](embed_mat1,embed_mat2)
                sim_score = pd.DataFrame(sim_score)
                sim_score.index   = tok1.values
                sim_score.columns = tok2.values
                if write_raw_pandas :st.write(sim_score,key = key)
                if sim_score.shape == (1,1) :
                    sim_score = sim_score.iloc[0][0]
                    if sim_score > threshold:
                        st.success(sim_score)
                        st.success(f'Scalar Similarity={sim_score} for distance metric={dist_algo}')
                        st.error('No similarity matrix for only 2 tokens. Try entering at least 1 sentences in a field')
                    else:
                        st.error(f'Scalar Similarity={sim_score} for distance metric={dist_algo}')
                else :
                    ploty_avaiable = False
                    try :
                        import plotly.express as px
                        ploty_avaiable = True
                    except :
                        ploty_avaiable = False
                    # for tok emb, sum rows and norm by rows, then sum cols and norm by cols to generate a scalar from matrix
                    scalar_sim_score  = np.sum((np.sum(sim_score,axis=0) / sim_score.shape[0])) / sim_score.shape[1]
                    if display_scalar_similarities:
                        if scalar_sim_score > threshold:st.success(f'Scalar Similarity :{scalar_sim_score} for distance metric={dist_algo}')
                        else: st.error(f'Scalar Similarity :{scalar_sim_score} for embedder={e_col} distance metric={dist_algo}')
                    if similarity_matrix:
                        if ploty_avaiable :
                            fig = px.imshow(sim_score)#, title=f'Simmilarity Matrix for embedding_model={e_name} distance metric={dist_algo}')
                            # st.write(fig,key =key)
                            similarity_metrics[f'{e_name}_{dist_algo}_similarity']={
                                'scalar_similarity' : scalar_sim_score,
                                'dist_metric' : dist_algo,
                                'embedding_model': e_name,
                                'modelhub_info' : modelhub_links[num_emb],
                             }

                            subh = f"""**Simmilarity Matrix:** Embedding-Model=`{e_name}`, Similarity-Score=`{scalar_sim_score}`,  distance metric=`{dist_algo}`"""
                            cols[col_index].markdown(subh)
                            cols[col_index].write(fig, key=key)

    display_similarity_summary = False # TODO ADD
    if display_similarity_summary:
        exp = st.beta_expander("Similarity summary")
        exp.write(similarity_metrics)
    if display_embed_information:
        exp = st.beta_expander("Embedding Vector Information")
        exp.write(embed_vector_info)




nlu.enable_streamlit_caching()
st.sidebar.write('lol') # elmo albert en.embed.bert.base_cased electra xlnet ar.embed.cbow biobert
display_word_similarity(nlu.load('glove '),('Peter likes to party', 'Angela likes to party'))


from sparknlp_jsl.annotator import ContextualParserApproach,ContextualParserModel

class ContextParser:
    @staticmethod
    def get_default_model():
        return ContextualParserApproach() \
            .setInputCols(["sentence", "token"]) \
            .setOutputCol("context_entity")

    @staticmethod
    def get_default_trainable_model():
        return ContextualParserApproach() \
            .setInputCols(["sentence", "token"]) \
            .setOutputCol("context_entity")



from sparknlp_jsl.annotator import DrugNormalizer

class DrugNorm:
    @staticmethod
    def get_default_model():
        return DrugNormalizer() \
            .setInputCols(["document"]) \
            .setOutputCol("normalized_drugs")




import os
import nlu
# use the generate_py_code_for_nlu_ref() method to convert a nlu_reference into a Spark NLP pipe


#every annotator class is accompanied with a component_infos.json
# in the same dir as the component_info there will be a .py
# Then just get the code inside of get_pretrained_model() and gg! (Basically just remove"return" and white spaces, gg?
# need to add parameters to .get_pretrained_model(name, language)
# need to update input/outut cols
# Some annotators dont have pretrained models and thus they have no get_pretrained_model() method. In these cases, we will inject the get_default_model() code and update no params
def generate_py_code_for_nlu_ref_or_pipe(nlu_ref = 'pos', pipe=None):
    pipe_code = []
    if pipe == None: pipe = nlu.load(nlu_ref)
    is_pretrained_pipe = None
    component_kind, nlp_ref,lang = get_component_metadata(nlu_ref)
    print(    pipe.predict('hello my name is fred'))
    if component_kind=='pipe' :
        pipe_code = make_pretrained_pipe_code(nlp_ref, lang,nlu_ref)
    else :
        for c in pipe.pipe_components:
            source_path = get_nlu_py_file_path_for_nlu_component(c)
            if check_if_has_pretrained(source_path) :
                pipe_code.append(extract_get_pretrained_source_code(source_path))
            else:
                pipe_code.append(extract_get_default_source_code(source_path))
        pipe_code = make_model_pipe_code_pretty(pipe_code,pipe, lang, nlp_ref,nlu_ref)
    print_py_pipe_code(pipe_code)
    return py_pipe_code_to_str(pipe_code)

def print_py_pipe_code(pipe_code):
    if isinstance(pipe_code[0],str): print('\n'.join(pipe_code))
    else:
        for l in pipe_code: print(''.join(l))

def py_pipe_code_to_str(pipe_code):
    if isinstance(pipe_code[0],str): return '\n'.join(pipe_code)
    else:
        return ''.join([item for sublist in pipe_code for item in sublist])
        #  return ''.join([l  for l in pipe_code])
        # for l in pipe_code: return ''.join(l)

def get_nlu_py_file_path_for_nlu_component(c):
# for a given NLU component, find source file that creates the Spark NLP annotators

    c_folder = c.component_path
    candidates = os.listdir(c_folder)
    py_file = ''
    for can in candidates:
        if'.py' in can and '__init__' not in can: py_file=can
    return c_folder + py_file


def check_if_has_pretrained(source_path):
    f = open(source_path, 'r')

    # check for a py file if it has a pretrained method or not
    for l in f : # check if annotator has .pretrained()
        if '.pretrained()' in l : return True
    return  False

def extract_get_default_source_code(source_path):
    #extract get_default() method source code
    f = open(source_path, 'r')
    source_lines = []
    func_start = False
    for l in f :
        if func_start and 'def' in l or l =='\n' and func_start : break # func we want to extract has ended
        if func_start: source_lines.append(clean_line(l))
        if 'get_default_model' in l : func_start = True
    return  source_lines

def extract_get_pretrained_source_code( source_path ):
    #extract get_pretrained() method source code

    f = open(source_path, 'r')

    source_lines = []
    func_start = False
    for l in f :
        if func_start and 'def' in l or l =='\n' and func_start: break # func we want to extract has ended
        if func_start: source_lines.append(clean_line(l))
        if 'get_pretrained_model' in l : func_start = True

    # todo remove \t of the FIRST line only!
    source_lines[0].replace('\t','')
    return  source_lines

def define_annotator_variable_names(pipe_code):
    # set left side of equality for variable name
    new_pipe = []
    var_names = []

    for anno in pipe_code:
        first_line = anno[0]
        class_name = first_line.split('=')[1].split('.')[0]\
            .replace(' ','').replace('\\','').\
            replace('(','').replace(')','')\
            .replace('name','').replace('language','')\
            .replace('\n', '')

        var_name = camel_case_to_underscore(class_name)
        var_names.append(var_name)
        new_pipe.append([anno[0].replace('annotator_name',var_name)] + anno[1:])
    return new_pipe,var_names

def camel_case_to_underscore(string):
    new_s = ''
    new_pipe =[]

    for i, s in enumerate(string):
        if i == 0 : new_s = s.lower()
        elif s.islower(): new_s = new_s+s
        else : new_s = new_s+ '_' +s.lower()
    return  new_s

def get_nlp_reference_for_nlu_reference(nlu_ref):
    # for NLU reference, search namesapce for matchng NLP eference
    pass

def configure_indentation(pipe_code):
    # first line should be  0 tab, all following lines should be 1  tab
    new_pipe =  []
    for anno in pipe_code:
        rem_white = lambda x : x.lstrip()
        prepend_indent = lambda x : '	' +x #we could use \t but it would not be renderd by jekyll(?)
        new_pipe.append( [anno[0].lstrip()] + list(map(prepend_indent,map(rem_white, anno[1:]))))

    return new_pipe

def configure_input_output_cols(pipe_code, nlu_pipeline):
    #inject paramets for input/output column into code base on what they are configured to in nlu pipeline
    new_pipe_code = []
    for i,anno in enumerate(pipe_code):
        input_col = nlu_pipeline.pipe_components[i].info.spark_input_column_names
        output_col = nlu_pipeline.pipe_components[i].info.spark_output_column_names
        new_anno_code = []
        for l in anno:
            new_l = ''
            if 'setOutputCols' in l:
                new_l = l[:l.find('(')+1] + str(output_col) +   l[l.find(')'): ] # ')'
            elif 'setOutputCol' in l:
                new_l = l[:l.find('(')+1] + str(output_col).replace('[','').replace(']','') +  l[l.find(')'): ] #  ')'
            elif 'setInputCols' in l:
                new_l = l[:l.find('(')+1] + str(input_col) +  l[l.find(')'): ] # ')'
            elif 'setInputCol' in l:
                new_l = l[:l.find('(')+1]+ str(input_col).replace('[','').replace(']','')  +  l[l.find(')'): ]
            else: new_l = l
            new_anno_code.append(new_l)

        new_pipe_code.append(new_anno_code)
    return new_pipe_code

def get_embedding_parameters(lang, nlp_storage_ref):
    #iterate over namespace and see if embed exist for that lang. if not, it must be multi lang embed
    # if embeddings do not exist in specific lang and xx space, it has to be en embedding..

    for k,v in nlu.Spellbook.pretrained_models_references[lang].items():
        if v == nlp_storage_ref : return lang, nlp_storage_ref

    for k,v in nlu.Spellbook.pretrained_models_references['xx'].items():
        if v == nlp_storage_ref : return 'xx', nlp_storage_ref

    for k,v in nlu.Spellbook.pretrained_models_references['en'].items():
        if v == nlp_storage_ref : return 'en', nlp_storage_ref

    print("EXCEPTION COULD NOT FIND CORRECT EMBED PARAMETERS FOR ", lang, nlp_storage_ref)
    return '???????', nlp_storage_ref

def get_pretrained_model_parameters(lang, nlu_model_ref):
    #iterate over namespace and see if embed exist for that lang. if not, it must be multi lang embed
    # if embeddings do not exist in specific lang and xx space, it has to be en embedding..

    for k,v in nlu.Spellbook.pretrained_models_references[lang].items():
        if k == nlu_model_ref : return lang, v

    for k,v in nlu.Spellbook.pretrained_models_references['xx'].items():
        if k == nlu_model_ref : return 'xx', v

    for k,v in nlu.Spellbook.pretrained_models_references['en'].items():
        if k == nlu_model_ref : return 'en', v

    print("EXCEPTION COULD NOT FIND CORRECT MODEL PARAMETERS FOR ", lang, nlu_model_ref)
    return '???????', nlu_model_ref


def configure_pretrained(pipe_code, nlu_pipeline, lang, nlp_ref, nlu_ref):
    #configure .pretrained parameters and also takes the embeddings into acound
    # We iterate in reverse, first will be classifier then model
    parameters_set_count = 0
    new_pipe_code = []
    for i,anno in enumerate(reversed(pipe_code)):

        if 'pretrained' in anno[0] and parameters_set_count != 0:
            storage_refs = [c.model.extractParamMap()[c.model.storageRef] for c in nlu_pipeline.pipe_components if any ([ 'storageRef' in str(d) for d in c.model.extractParamMap().keys() ] ) ]
            p_lang,p_name = get_embedding_parameters(lang,storage_refs[0])

            new_l = anno[0][:anno[0].find('(') + 1] +"'"+ p_name +"'"+  ', ' +"'"+  p_lang +"'"+ anno[0][anno[0].find(')'):]
            parameters_set_count =+1
            new_pipe_code.append([new_l] + anno[1:])

        elif 'pretrained' in anno[0] and parameters_set_count== 0:
            p_name, p_lang = get_pretrained_model_parameters(lang,nlu_ref)
            parameters_set_count = 1
            new_l = anno[0][:anno[0].find('(') + 1] +"'"+ p_name +"'"+ ', ' +"'"+  p_lang +"'"+ anno[0][anno[0].find(')'):]
            new_pipe_code.append([new_l] + anno[1:])
        else:
            new_pipe_code.append(anno)

    if parameters_set_count > 2 : print("CONFIGURED MROE THAn 2 PARAMS FOR ", nlu_ref,nlp_ref, nlu_pipeline, pipe_code)
    new_pipe_code.reverse()
    return  new_pipe_code

def configure_annotator_parameters(pipe_code, nlu_pipeline, lang, nlp_ref,nlu_ref):
    # 1. inject parameters  of input/output cols, based on what NLU has currently configured them to
    # 2. If .pretrained(name,lang) in first line, inject name/lang
    # (2.) occurs only for NON pretrained pipelines. We also need to check if the model uses embeddings and if so, we must configure .pretraiend() of the embeddings aswell
    pipe_code = configure_input_output_cols(pipe_code, nlu_pipeline)
    pipe_code = configure_pretrained(pipe_code, nlu_pipeline, lang, nlp_ref,nlu_ref)
    return pipe_code



def clean_line(l): return l.replace('return','annotator_name = ').replace('\t\t','\t')


def make_model_pipe_code_pretty(pipe_code, nlu_pipeline, lang, nlp_ref,nlu_ref):
    # pipe code consist of list of lists. each outer list contains one Annotator Source code. Each inner element of a list is a list with strings that represent the source code
    # we replace annotator_name with a more meaningful name
    # also we fix tabs
    # inject parameters aswell input/output cols and .pretrained(name,lang)
    pipe_code,var_names = define_annotator_variable_names(pipe_code)
    pipe_code = configure_indentation(pipe_code)

    # set input/output cols and .pretrained() parameters
    pipe_code = configure_annotator_parameters(pipe_code, nlu_pipeline, lang, nlp_ref,nlu_ref)

    #pipeline with stages definition
    pipe_code.append(get_spark_pipe_definition(var_names))
    pipe_code.insert(0,[add_pretrained_model_imports()])
    # fit_transform and dataframe definition
    pipe_code.append([get_fit_transform_df_suffix()])



    print("DONE MAKING PRETTY")

    return pipe_code
def make_pretrained_pipe_code(nlp_ref, lang,nlu_ref):
    pipe_code = []
    pipe_code.append(add_pretrained_model_imports())
    pipe_code.append(add_pretrained_pipe_definition(nlp_ref,lang))
    pipe_code.append(get_fit_transform_df_suffix())
    return pipe_code



def add_pretrained_pipe_imports(): return 'import sparknlp \nfrom sparknlp.pretrained\nimport PretrainedPipeline\n'
def add_pretrained_model_imports(): return 'import sparknlp \nfrom sparknlp.annotator import *\nfrom pyspark.ml import Pipeline\n'

def add_pretrained_pipe_definition(nlp_ref,lang): return f"pipeline = PretrainedPipeline('{nlp_ref}', lang='{lang}')\n"


def get_spark_pipe_definition(var_names):
    # summarize all annotators in a Spark pipeline  and add it as new line to the list
    #var_name is a list of what the annotator variable namesa re in the code

    pre = "nlp_pipe = Pipeline(stages=["
    stages = ','.join(var_names)
    post = "])\n"
    return pre+stages+post


#todo add get_sample_df
def get_fit_transform_df_suffix():
    # after we have the full model/pretraine pipe code and the 'pipe' variable, we can always add the same suffix!
    l1 = "text = 'John Snow Labs offers over 300 models in over 40 languages to solve a variety of NLP problems while keeping the highest accuracy.'\n"
    l2 = "df = spark.createDataFrame([[text]]).toDF('text')\n"
    return l1+l2+'predictions = nlp_pipe.fit(df).transform(df)\npredictions.printSchema()\npredictions.show()'


def get_component_metadata(nlu_reference):
# get nlp_ref and wheter it is pretrained pipe or just a prertained mdoel(component kind)
    from nlu import all_components_info
    infos = nlu_reference.split('.')
    language = ''
    component_type = ''
    dataset = ''
    component_embeddings = ''
    component_pipe = []
    if len(infos) == 0: 1+1
    elif len(infos) == 1:
        # if we only have 1 split result, it must a a NLU action reference or an alias
        language = 'en'
        if infos[0] in all_components_info.all_components or all_components_info.all_nlu_actions:
            component_type = infos[0]
    #  check if it is any query of style #<lang>.<class>.<dataset>.<embeddings>
    elif infos[0] in all_components_info.all_languages:
        language = infos[0]
        component_type = infos[1]

        if len(infos) == 3:  # dataset specified
            dataset = infos[2]
        if len(infos) == 4:  # embeddings specified
            component_embeddings = infos[3]

    # passing embed_sentence can have format embed_sentence.lang.embedding or embed_sentence.embedding
    # i.e. embed_sentence.bert
    # fr.embed_sentence.bert will automatically select french bert thus no embed_sentence.en.bert or simmilar is required
    # embed_sentence.bert or en.embed_sentence.bert
    # name does not start with a language
    # so query has format <class>.<dataset>
    elif len(infos) == 2:
        language = 'en'
        component_type = infos[0]
        dataset = infos[1]
    # query has format <class>.<dataset>.<embeddings>
    elif len(infos) == 3:
        language = 'en'
        component_type = infos[0]
        dataset = infos[1]
        component_embeddings = infos[1]

    resolved_component = resolve_component_from_parsed_query_data(language, component_type, dataset,
                                                                  component_embeddings, nlu_reference)

    return resolved_component
def resolve_component_from_parsed_query_data(language, component_type, dataset, component_embeddings, nlu_ref):
    from nlu import Spellbook
    component_kind = ''  # either model or pipe or auto_pipe
    nlp_ref = ''
    resolved = False
    # 1. check if pipeline references for resolution
    if resolved == False and language in Spellbook.pretrained_pipe_references.keys():
        if nlu_ref in Spellbook.pretrained_pipe_references[language].keys():
            component_kind = 'pipe'
            nlp_ref = Spellbook.pretrained_pipe_references[language][nlu_ref]
            resolved = True

    # 2. check if model references for resolution
    if resolved == False and language in Spellbook.pretrained_models_references.keys():
        if nlu_ref in Spellbook.pretrained_models_references[language].keys():
            component_kind = 'model'
            nlp_ref = Spellbook.pretrained_models_references[language][nlu_ref]
            resolved = True

    return (component_kind, nlp_ref,language)


if __name__ == "__main__":
    print('___________________PIPE TEST ___________________________-')
    pipe_code = generate_py_code_for_nlu_ref_or_pipe('albert')
    print("DONE CODE : ", pipe_code)
    # print(pipe_code)

    print('___________________COMPONENT TEST ___________________________-')
    # should proeprly set .pretrained prameters of embdeddings and model
    pipe_code = generate_py_code_for_nlu_ref_or_pipe('nl.ner.wikiner.glove.6B_100')
    print("DONE CODE : ", pipe_code)

#todo
def generate_jupyter_notebook():

    pass


from nlu.pipe.pipe_components import SparkNLUComponent


class Embeddings(SparkNLUComponent):

    def __init__(self, annotator_class='glove', lang ='en', component_type='embedding', get_default=True, model = None, nlp_ref ='', nlu_ref ='', is_licensed=False, resolution_ref='',loaded_from_pretrained_pipe=False,do_ref_checks=True ):
        if do_ref_checks:
            if 'use' in nlu_ref or 'tfhub_use' in nlp_ref: annotator_class = 'use'
            # first check for sentence then token embeddings.

            elif 'xlm'     in nlu_ref or 'xlm' in nlp_ref : annotator_class = 'xlm'
            elif 'roberta' in nlu_ref or 'roberta' in nlp_ref : annotator_class = 'roberta'
            elif 'distil'  in nlu_ref or 'distil' in nlp_ref : annotator_class = 'distil_bert'


            elif 'bert' in nlp_ref and 'albert' not in nlp_ref and 'sent' in nlp_ref : annotator_class= 'sentence_bert'
            elif 'bert' in nlu_ref and 'albert' not in nlu_ref and 'sent' in nlu_ref : annotator_class= 'sentence_bert'

            elif 'elmo' in nlp_ref  : annotator_class= 'elmo'
            elif 'elmo' in nlu_ref  : annotator_class= 'elmo'


            elif 'electra' in nlp_ref and 'sent' in nlp_ref : annotator_class= 'sentence_bert'
            elif 'electra' in nlu_ref and 'sent' in nlu_ref : annotator_class= 'sentence_bert'

            elif 'bert' in nlu_ref and 'albert' not in nlu_ref: annotator_class= 'bert'
            elif 'bert' in nlp_ref and 'albert' not in nlp_ref: annotator_class= 'bert'

            elif 'electra' in nlu_ref or 'electra' in nlp_ref:  annotator_class= 'bert'
            elif 'labse' in nlu_ref   or 'labse' in nlp_ref:  annotator_class= 'sentence_bert'

            elif 'tfhub' in nlu_ref or 'tfhub' in nlp_ref: annotator_class= 'use'
            elif 'glove' in nlu_ref or 'glove' in nlp_ref : annotator_class = 'glove'
            elif 'cc_300d' in nlu_ref or 'cc_300d' in nlp_ref : annotator_class = 'glove'

            elif 'albert' in nlu_ref or 'albert' in nlp_ref : annotator_class = 'albert'
            elif 'xlnet' in nlu_ref or 'xlnet' in nlp_ref : annotator_class = 'xlnet'

            # Default component models for nlu actions that dont specify a particular model
            elif 'embed_sentence' in nlu_ref : annotator_class = 'glove'
            elif 'embed' in nlu_ref          : annotator_class = 'glove'

        if model != None : self.model = model
        else :

            # Check if this lang has embeddings, if NOT set to multi lang xx!
            multi_lang_embeds = ['th']
            if lang in multi_lang_embeds : lang ='xx'

            if 'xlm' in annotator_class :
                from nlu import XLM
                if get_default: self.model =  XLM.get_default_model()
                else : self.model = XLM.get_pretrained_model(nlp_ref, lang)

            elif 'roberta' in annotator_class :
                from nlu import Roberta
                if get_default: self.model =  Roberta.get_default_model()
                else : self.model = Roberta.get_pretrained_model(nlp_ref, lang)


            elif 'distil_bert' in annotator_class :
                from nlu import DistilBert
                if get_default: self.model =  DistilBert.get_default_model()
                else : self.model = DistilBert.get_pretrained_model(nlp_ref, lang)

            elif 'albert' in annotator_class :
                from nlu import SparkNLPAlbert
                if get_default: self.model =  SparkNLPAlbert.get_default_model()
                else : self.model = SparkNLPAlbert.get_pretrained_model(nlp_ref, lang)
            elif 'bert' in annotator_class and 'sent' in annotator_class  :
                from nlu import BertSentence
                if get_default : self.model =  BertSentence.get_default_model()
                elif is_licensed : self.model = BertSentence.get_pretrained_model(nlp_ref, lang,'clinical/models' )
                else : self.model = BertSentence.get_pretrained_model(nlp_ref, lang)
            elif 'electra' in annotator_class and 'sent' in annotator_class  :
                from nlu import BertSentence
                if get_default : self.model =  BertSentence.get_default_model()
                elif is_licensed : self.model = BertSentence.get_pretrained_model(nlp_ref, lang,'clinical/models' )
                else : self.model = BertSentence.get_pretrained_model(nlp_ref, lang)
            elif 'bert' in annotator_class :
                from nlu import SparkNLPBert
                if get_default : self.model =  SparkNLPBert.get_default_model()
                elif is_licensed : self.model = SparkNLPBert.get_pretrained_model(nlp_ref, lang,'clinical/models' )
                else : self.model = SparkNLPBert.get_pretrained_model(nlp_ref, lang)
            elif 'elmo' in annotator_class  :
                from nlu import SparkNLPElmo
                if get_default : self.model = SparkNLPElmo.get_default_model()
                else : self.model =SparkNLPElmo.get_pretrained_model(nlp_ref, lang)
            elif  'xlnet' in annotator_class  :
                from nlu import SparkNLPXlnet
                if get_default : self.model = SparkNLPXlnet.get_default_model()
                else : self.model = SparkNLPXlnet.get_pretrained_model(nlp_ref, lang)
            elif 'use' in annotator_class   :
                from nlu import SparkNLPUse
                if get_default : self.model = SparkNLPUse.get_default_model()
                else : self.model = SparkNLPUse.get_pretrained_model(nlp_ref, lang)
            elif 'glove' in annotator_class   :
                from nlu import Glove
                if annotator_class == 'glove' and get_default==True: self.model = Glove.get_default_model()
                else :
                    if get_default : self.model = Glove.get_default_model()
                    elif is_licensed : self.model = Glove.get_pretrained_model(nlp_ref,lang,'clinical/models')
                    else :
                        if nlp_ref == 'glove_840B_300' or  nlp_ref== 'glove_6B_300':
                            # if lang=='en' and nlp_ref=='glove_6B_300': #special case
                            lang = 'xx' # For these particular Glove embeddings, anyreference to them is actually the reference to the multilingual onces
                            self.model = Glove.get_pretrained_model(nlp_ref, lang)
                        else :
                            self.model = Glove.get_pretrained_model(nlp_ref, lang)

        SparkNLUComponent.__init__(self, annotator_class, component_type,nlu_ref,nlp_ref,lang)



import requests

url = 'https://nlp.johnsnowlabs.com/docs/en/annotators'
r  = requests.get(url)
from bs4 import BeautifulSoup

data = r.text
soup = BeautifulSoup(data,features="html.parser")
print (soup.head.next_element)

for header in soup.find_all("div", {"class": "h3-box"}):
    # print(header)
    para = header.findNext('p')
    print(para.g)
    # print(header.next_siblig)



def df_to_pretty_markdown(df,max_len =  50):
    '''
    Truncates every cell content to max len and then calls .to_markdown() on the table
    Especially usefull for embeddings, which are very long.
    :param df: pandas dataframe to convert
    :param max_len: max cell len when converter to str.
    :return:
    '''
    print('wring Markdown!')
    for idx, row in df.iterrows():
        for c in df.columns:
            cell = str(df.iloc[idx][c])
            if len(str(cell)) > max_len :
                df.loc[idx][c] = cell[:max_len] + '..]'
    md = df.to_markdown()
    print('wring Markdown!',md)
    return md

# models_df = pd.read_csv('generated_data/models_generated_with_meta.csv')
def test_every_default_component(nlu_reference, id,pretrained_model_cache_path = '/home/loan/cache_pretrained'):
    import os
    from os import path
    import pandas as pd


    # __file__ should be defined in this case
    PARENT_DIR = path.dirname(path.dirname(path.abspath(__file__)))
    sys.path.append(PARENT_DIR)
    conf_path = os.getcwd()
    sys.path.append(conf_path)
    #rm  old models to safe us from full hdd
    PARENT_DIR = path.dirname(path.dirname(path.abspath(__file__)))
    sys.path.append(PARENT_DIR)
    conf_path = os.getcwd()
    sys.path.append(conf_path)
    sys.path.append('/home/loan/Documents/freelancework/jsl/nlu/4realnlugit/tmp/doc_generation')
    if path.exists (os.path.expanduser(pretrained_model_cache_path)): os.system('rm -r ' + pretrained_model_cache_path)

    import generate_pipe_code as pipe_code_generator
    import nlu

    import gc
    id = int(id)
    print('testing', nlu_reference)
    if 'NOT FOUND' in nlu_reference :
        print('Skipping for nlu ref with id ', id)
        return

    string_to_predict = 'John Snow Labs offers over 300 models in over 40 languages to solve a variety of NLP problems while keeping the highest accuracy.'

    models_df = pd.read_csv('./tmp/doc_generation/generated_data/models_generated_with_meta.csv')

    print(f'Test NLU_REF={nlu_reference}')
    nlu_component_pipe = nlu.load(nlu_reference,verbose=True)
    nlu_component = nlu.parse_component_data_from_name_query(nlu_reference)
    if type (nlu_component ) is type([]):
        #pipe handling
        models_df.at[id, 'model_class'] =  'PretrainedPipeline'
        models_df.at[id, 'model_upstream_deps'] = [c.model.extractParamMap()[c.model.storageRef] for c in nlu_component_pipe.pipe_components if any ([ 'storageRef' in str(d) for d in c.model.extractParamMap().keys() ] ) ] # [c.model.extractParamMap()[c.model.storageRef] for c in nlu_component_pipe.pipe_components] #'Pipe storage dependencies already met']
        models_df.at[id, 'consists_of'] = [m.model.__class__.name for m in nlu_component]
        models_df.at[id, 'can_be_satisfied_with_classes'] = [c.model.__class__.name for c in nlu_component_pipe.pipe_components]
        models_df.at[id, 'type'] = 'pipeline'
        models_df.at[id, 'nlu_type'] = 'pipeline'
        models_df.at[id, 'nlu_input_columns'] = [c.info.inputs for c in nlu_component_pipe.pipe_components]
        models_df.at[id, 'nlu_output_columns'] = [c.info.outputs for c in nlu_component_pipe.pipe_components]
        models_df.at[id, 'nlp_input_columns'] = [c.info.spark_input_column_names for c in nlu_component_pipe.pipe_components]
        models_df.at[id, 'nlp_output_columns'] =  [c.info.spark_output_column_names for c in nlu_component_pipe.pipe_components]
        models_df.at[id, 'nlu_predictions'] = nlu_component_pipe.predict(string_to_predict, metadata=True, positions=True, output_level='document', drop_irrelevant_cols=False)
        models_df.at[id, 'nlu_predictions_markdown'] = df_to_pretty_markdown(nlu_component_pipe.predict(string_to_predict, metadata=True, positions=True, output_level='document', drop_irrelevant_cols=False))
        models_df.at[id, 'nlp_pipeline_code'] = pipe_code_generator.generate_py_code_for_nlu_ref_or_pipe(nlu_ref,nlu_component_pipe)

    else :
        # get metdata for single model
        models_df.at[id, 'model_class'] =  nlu_component.model.__class__.name
        if any ([ 'storageRef' in str(d) for d in nlu_component.model.extractParamMap().keys() ] ) : models_df.at[id, 'model_upstream_deps'] = nlu_component.model.extractParamMap()[nlu_component.model.storageRef]
        else : models_df.at[id, 'model_upstream_deps'] = "Has no upstream dependencies"
        models_df.at[id, 'consists_of'] =  nlu_component.model.__class__.name
        models_df.at[id, 'can_be_satisfied_with_classes'] =  [c.model.__class__.name for c in nlu_component_pipe.pipe_components]
        models_df.at[id, 'type'] = 'model'
        models_df.at[id, 'nlu_type'] = nlu_component.info.type
        models_df.at[id, 'nlu_input_columns'] = nlu_component.info.inputs
        models_df.at[id, 'nlu_output_columns'] = nlu_component.info.outputs
        models_df.at[id, 'nlp_input_columns'] = nlu_component.info.spark_input_column_names
        models_df.at[id, 'nlp_output_columns'] = nlu_component.info.spark_output_column_names
        models_df.at[id, 'nlu_predictions'] = nlu_component_pipe.predict(string_to_predict,metadata=True, positions=True, output_level='document', drop_irrelevant_cols=False)
        models_df.at[id, 'nlu_predictions_markdown'] = df_to_pretty_markdown(nlu_component_pipe.predict(string_to_predict,metadata=True, positions=True, output_level='document', drop_irrelevant_cols=False))
        models_df.at[id, 'nlp_pipeline_code'] = pipe_code_generator.generate_py_code_for_nlu_ref_or_pipe(nlu_ref,nlu_component_pipe)


    models_df.to_csv('./tmp/doc_generation/generated_data/models_generated_with_meta.csv',index=False)

    print('TESTING DONE FOR NLU REFERENCE : ', nlu_reference)







import sys
print ('Number of arguments:', len(sys.argv), 'arguments.')
print ('Argument List:', str(sys.argv))
nlu_ref = sys.argv[1]
id = sys.argv[2]
pretrained_model_cache_path =  sys.argv[3]
test_every_default_component(nlu_ref,id,pretrained_model_cache_path)



import pandas as pd
import nlu
from typing import List, Sequence, Tuple, Optional, Dict, Union, Callable
import streamlit as st
def get_pipe(model='ner'): return nlu.load(model)
def display_side_ui(): pass

def viz_streamlit(
    # Base Params
    default_text:Union[str, List[str], pd.DataFrame, pd.Series],
    model_selection:List[str],
    # NER PARAMS
    default_ner_model2viz:Union[str, List[str]] = 'en.ner.onto.electra.base',

    # SIMILARITY PARAMS
    similarity_texts:Tuple[str,str]= ('Donald Trump Likes to part', 'Angela Merkel likes to party'),
    title:str = 'NLU ❤️ Streamlit - Prototype your NLP startup in 0 lines of code' ,
    # UI PARAMS
    visualizers:List[str] = ( "dependency_tree", "ner",  "similarity", "token_information", "embedding_word",'classification','embedding_sentence'), #viz2apply(#
    show_models_info:bool = True,
    show_model_select:bool = True,
    show_viz_selection:bool = True,
    show_logo:bool=True,
    show_code_snippets:bool=False,
    )->None:
    """Visualize either individual building blocks for streamlit or a full UI to experiment and explore models with"""
    from nlu.pipe.viz.streamlit_viz.streamlit_viz_tracker import StreamlitVizTracker

    if title: st.title(title)
    if show_logo :StreamlitVizTracker.show_logo()
    # Join list of NLU refs to str
    if isinstance(default_ner_model2viz, List)       : default_ner_model2viz = ' '.join(default_ner_model2viz)
    ner_model_2_viz     = default_ner_model2viz
    default_text    = st.text_area("Enter text you want to visualize below", default_text)
    if show_model_select :
        st.sidebar.checkbox('Show generated code samples')
        ner_model_2_viz = st.sidebar.selectbox("Select a NER model",model_selection,index=model_selection.index(default_ner_model2viz.split(' ')[0]),)
    active_visualizers = visualizers
    if show_viz_selection: active_visualizers = st.sidebar.multiselect("Visualizers",options=visualizers,default=visualizers)

    all_models = ner_model_2_viz + ' en.dep.typed '
    if show_models_info : StreamlitVizTracker.display_model_info(all_models)

    if 'dependency_tree' in active_visualizers :
        tree_pipe = get_pipe('en.dep.typed')
        StreamlitVizTracker.visualize_dep_tree(tree_pipe, default_text)
    if 'ner' in active_visualizers :
        ner_pipe = get_pipe(ner_model_2_viz)
        StreamlitVizTracker.visualize_ner(ner_pipe, default_text)
    if 'token_information' in active_visualizers:
        ner_pipe = get_pipe(ner_model_2_viz)
        StreamlitVizTracker.visualize_tokens_information(ner_pipe, default_text)
    if 'classification' in active_visualizers:
        ner_pipe = get_pipe(ner_model_2_viz+' sentiment')
        StreamlitVizTracker.visualize_classes(ner_pipe, default_text)
    if 'similarity' in active_visualizers:
        ner_pipe = get_pipe(ner_model_2_viz)
        StreamlitVizTracker.display_word_similarity(ner_pipe, similarity_texts)




def visualize_IU():
    """Viz NLU UI"""
    # Get a list of all NER models and define default input text
    ner_models = nlu.get_components('ner',include_pipes=True)
    default_text = "Sundar Pichai is the CEO of Google."
    viz_streamlit(default_text,ner_models)
visualize_IU()












from sparknlp_jsl.annotator import SentenceEntityResolverModel,SentenceEntityResolverApproach

class SentenceResolver:
    @staticmethod
    def get_pretrained_model(name, language, bucket='clinical/models'):
        return SentenceEntityResolverModel.pretrained(name, language,bucket) \
            .setInputCols(["sentence_embeddings"]) \
            .setDistanceFunction("COSINE") \
            .setNeighbours(3) \
            .setOutputCol("sentence_resolution")

    @staticmethod
    def get_default_trainable_model():
        return SentenceEntityResolverApproach() \
        .setInputCols("sentence_embeddings") \
        .setOutputCol("sentence_resolution") \
        .setDistanceFunction("COSINE") \
        .setNeighbours(1) \
        .setLabelCol('label')



from sparknlp.annotator import *

class Glove:
    @staticmethod
    def get_default_model():
        return WordEmbeddingsModel.pretrained() \
        .setInputCols("sentence", "token") \
        .setOutputCol("word_embeddings")


    @staticmethod
    def get_pretrained_model(name, language, bucket = None):
        return WordEmbeddingsModel.pretrained(name,language,bucket) \
            .setInputCols("sentence", "token") \
            .setOutputCol("word_embeddings")









import unittest
from nlu import *

class TestEmotion(unittest.TestCase):

    def test_emotion_model(self):
        # NLU will predict both as happy. If you reverse order both become sad

        pipe = nlu.load('emotion',verbose=True)
        # df = pipe.predict(['I love pancaces. I hate Mondays', 'I love Fridays'], output_level='sentence',drop_irrelevant_cols=False, metadata=True, )
        # for c in df.columns: print(df[c])
        df = pipe.predict(['I love pancaces. I hate Mondays', 'I love Fridays'], output_level='document')
        for c in df.columns: print(df[c])


if __name__ == '__main__':
    unittest.main()




"""
This module contains methods that will be applied inside of the apply call on the freshly convertet
Spark DF that is now as list of dicts.

These methods are meant to be applied in the internal calls of the metadata extraction.
They expect dictionaries which represent the metadata field extracted from Spark NLP annotators.


"""
import numpy as np
def meta_extract_language_classifier_max_confidence(row,configs):
    ''' Extract the language classificationw ith highest confidence and drop the others '''
    # # todo Get the best, but what about TOP K!  conditional sentence extraction and mroe docs
    #unpack all confidences to float and set 'sentence' key value to -1 so it does not affect finding the highest cnfidence
    unpack_dict_values = lambda x : -1 if 'sentence' in x[0]  else float(x[1][0])
    l = list(map(unpack_dict_values,row.items()))
    m = np.argmax(l)
    k = list(row.keys())[m]

    return {k+'_confidence' : row[k][0]} # remoe [0] for list return

def zipp(l): return zip(*l) # unpack during list comprehension not supported in Python, need this workaround for now

def extract_maximum_confidence(row, configs):
    ''' Extract the maximum confidence from any classifier with N classes.
    A classifier with N classes, has N confidences in it's metadata by default, which is too much data usually.
    This extractor gets the highest confidence from the array of confidences.
    This method assumes all keys in metadata corrospond to confidences, except the `sentence` key, which maps to a sentence ID
    key schema is 'meta_' + configs.output_col_prefix + '_confidence'

    Parameters
    -------------
    configs : SparkNLPExtractorConfig
    if configs.get_sentence_origin is True, the sentence origin column will be kept, otherwise dropped.
    row : dict
        i.e. looks like{'meta_category_sentence': ['0'],'meta_category_surprise': ['0.0050183665'],'meta_category_sadness': ['8.706827E-5'],'meta_category_joy': ['0.9947379'],'meta_category_fear': ['1.5667251E-4']}
    Returns
    ------------
    dict
      if configs.get_sentence_origin True  {'meta_sentiment_dl_sentence': ['0', '1'], 'meta_sentiment_dl_confidence': [0.9366506, 0.9366506]}
      else {'meta_sentiment_dl_confidence': [0.9366506, 0.9366506]}
    '''
    meta_sent_key = 'meta_' + configs.output_col_prefix + '_sentence'
    fl = lambda k : False if 'sentence' in k else True # every key  that has not the sub string sentence in it is considerd a confidence key
    confidences_keys = list(filter (fl, row.keys()))

    if configs.pop_meta_list :
        return {
            **{'meta_' + configs.output_col_prefix + '_confidence':max([float(row[k][0])  for k in confidences_keys ])},
            **({'meta_' + configs.output_col_prefix + '_sentence' : row[meta_sent_key]} if configs.get_sentence_origin else {})
        }
    else:
        if len(confidences_keys) == 1 :
            return {
                **{'meta_' + configs.output_col_prefix + '_confidence':max([float(row[k])  for k in confidences_keys ])},
                **({'meta_' + configs.output_col_prefix + '_sentence' : row[meta_sent_key]} if configs.get_sentence_origin else {})
            }
        else :
            return {
                **{'meta_' + configs.output_col_prefix + '_confidence': [ max(z )for z in zipp (list(map(float,row[k])) for k in confidences_keys)]} ,
                **({'meta_' + configs.output_col_prefix + '_sentence' : row[meta_sent_key]} if configs.get_sentence_origin else {})
            }




BASE_CODES = {
'SIMILARITY':"""
nlu.load({}).viz_streamlit_word_similarity(['{}', '{}'])""",

'TOKEN':"""
nlu.load({}).viz_streamlit_token({})""",

'TREE':"""
nlu.load({}).viz_streamlit_dep_tree({})""",

'NER':"""
nlu.load({}).viz_streamlit_ner({})""",

'CLASSES':"""
nlu.load({}).viz_streamlit_classes({})""",

'FULL_UI':"""
????
""",


'PREFIX': """
import nlu
""",

'INSTALL':"""
[Refer to NLU install docs for installing NLU](https://nlu.johnsnowlabs.com/docs/en/install)
""",

}
from  typing import List,Union
def get_code_for_viz(viz_type:str,models:Union[str, List[str]],data:Union[str,List[str]],max_len:int=100)->str:
    """Generate code sample for displaying how to generate visualization"""
    if viz_type =='SIMILARITY':
        if len(data) >= 2 and isinstance(data[0],str) and isinstance(data[0],str):
            models = f"'{' '.join(models)}'"
            return BASE_CODES[viz_type].format(models, data[0][:max_len], data[1][:max_len])
    else :
        models = f"'{models}'" if isinstance(models, str)  else f"'{' '.join(models)}'"
        data = f"'{data[:max_len]}'"
        return  BASE_CODES[viz_type].format(models, data)

